"""Custom exceptions for the vcs2l package."""

import sys


class Vcs2lError(Exception):
    """Base exception for vcs2l-specific errors."""

    def __init__(self, message: str):
        self.message = message
        super().__init__(message)


class UnsupportedPythonVersionError(Vcs2lError):
    """Raised when the Python version is too old for vcs2l."""

    def __init__(self, min_version: str = '3.6'):
        current_version = f'{sys.version_info.major}.{sys.version_info.minor}'
        message = (
            f'Unsupported Python version ({current_version}). '
            f'vcs2l requires Python {min_version} or higher.'
        )
        super().__init__(message)



import argparse
import os
from multiprocessing import cpu_count

from vcs2l.crawler import find_repositories
from vcs2l.executor import (
    execute_jobs,
    generate_jobs,
    output_repositories,
    output_results,
)


class Command(object):
    command = None

    def __init__(self, args):
        self.debug = args.debug if 'debug' in args else False
        self.hide_empty = args.hide_empty if 'hide_empty' in args else False
        self.nested = args.nested if 'nested' in args else False
        self.output_repos = args.repos if 'repos' in args else False
        if 'paths' in args:
            self.paths = args.paths
        else:
            self.paths = [args.path]


def check_greater_zero(value):
    try:
        value = int(value)
    except ValueError:
        raise argparse.ArgumentTypeError("invalid int value: '%s'" % value)
    if value <= 0:
        raise argparse.ArgumentTypeError("invalid positive int value: '%d'" % value)
    return value


def add_common_arguments(
    parser, skip_hide_empty=False, skip_nested=False, path_nargs='*', path_help=None
):
    parser.formatter_class = argparse.ArgumentDefaultsHelpFormatter
    group = parser.add_argument_group('Common parameters')
    group.add_argument(
        '--debug', action='store_true', default=False, help='Show debug messages'
    )
    if not skip_hide_empty:
        group.add_argument(
            '-s',
            '--hide-empty',
            '--skip-empty',
            action='store_true',
            default=False,
            help='Hide repositories with empty output',
        )
    if not skip_nested:
        group.add_argument(
            '-n',
            '--nested',
            action='store_true',
            default=False,
            help='Search for nested repositories',
        )
    try:
        default_workers = cpu_count()
    except NotImplementedError:
        default_workers = 4
    group.add_argument(
        '-w',
        '--workers',
        type=check_greater_zero,
        metavar='N',
        default=default_workers,
        help='Number of parallel worker threads',
    )
    group.add_argument(
        '--repos',
        action='store_true',
        default=False,
        help='List repositories which the command operates on',
    )
    if path_nargs == '?':
        path_help = path_help or 'Base path to look for repositories'
        group.add_argument(
            'path',
            nargs=path_nargs,
            type=existing_dir,
            default=os.curdir,
            help=path_help,
        )
    elif path_nargs == '*':
        path_help = path_help or 'Base paths to look for repositories'
        group.add_argument(
            'paths',
            nargs=path_nargs,
            type=existing_dir,
            default=[os.curdir],
            help=path_help,
        )


def existing_dir(path):
    if not os.path.exists(path):
        raise argparse.ArgumentTypeError("Path '%s' does not exist." % path)
    if not os.path.isdir(path):
        raise argparse.ArgumentTypeError("Path '%s' is not a directory." % path)
    return path


def simple_main(parser, command_class, args=None):
    add_common_arguments(parser)
    args = parser.parse_args(args)

    command = command_class(args)
    clients = find_repositories(command.paths, nested=command.nested)
    if command.output_repos:
        output_repositories(clients)
    jobs = generate_jobs(clients, command)
    results = execute_jobs(
        jobs, show_progress=True, number_of_workers=args.workers, debug_jobs=args.debug
    )

    output_results(results, hide_empty=args.hide_empty)

    any_error = any(r['returncode'] for r in results)
    return 1 if any_error else 0



import argparse
import sys

from vcs2l.clients import vcs2l_clients
from vcs2l.commands.command import Command, add_common_arguments
from vcs2l.commands.import_ import get_repositories
from vcs2l.executor import ansi, execute_jobs, output_results
from vcs2l.streams import set_streams


class ValidateCommand(Command):
    command = 'validate'
    help = 'Validate the repository list file'

    def __init__(self, args, url, version=None):
        super(ValidateCommand, self).__init__(args)
        self.url = url
        self.version = version
        self.retry = args.retry


def get_parser():
    parser = argparse.ArgumentParser(
        description='Validate a repositories file', prog='vcs validate'
    )
    group = parser.add_argument_group('"validate" command parameters')
    group.add_argument('--input', type=argparse.FileType('r'), default='-')
    group.add_argument(
        '--retry',
        type=int,
        metavar='N',
        default=2,
        help='Retry commands requiring network access N times on failure',
    )
    return parser


def generate_jobs(repos, args):
    jobs = []
    for path, repo in repos.items():
        clients = [c for c in vcs2l_clients if c.type == repo['type']]
        if not clients:
            from vcs2l.clients.none import NoneClient

            job = {
                'client': NoneClient(path),
                'command': None,
                'cwd': path,
                'output': "Repository type '%s' is not supported" % repo['type'],
                'returncode': NotImplemented,
            }
            jobs.append(job)
            continue

        client = clients[0](path)
        args.path = None  # expected to be present
        command = ValidateCommand(
            args, repo['url'], str(repo['version']) if 'version' in repo else None
        )
        job = {'client': client, 'command': command}
        jobs.append(job)
    return jobs


def main(args=None, stdout=None, stderr=None):
    set_streams(stdout=stdout, stderr=stderr)

    parser = get_parser()
    add_common_arguments(parser, skip_nested=True, path_nargs=False)
    args = parser.parse_args(args)
    try:
        repos = get_repositories(args.input)
    except RuntimeError as e:
        print(ansi('redf') + str(e) + ansi('reset'), file=sys.stderr)
        return 1

    jobs = generate_jobs(repos, args)

    results = execute_jobs(
        jobs, show_progress=True, number_of_workers=args.workers, debug_jobs=args.debug
    )

    output_results(results, hide_empty=args.hide_empty)

    any_error = any(r['returncode'] for r in results)
    return 1 if any_error else 0


if __name__ == '__main__':
    sys.exit(main())



import sys

stdout = sys.stdout
stderr = sys.stderr


def set_streams(stdout=None, stderr=None):
    _set_streams(stdout_=stdout, stderr_=stderr)


def _set_streams(stdout_=None, stderr_=None):
    global stdout
    global stderr
    if stdout_ is not None:
        stdout = stdout_
    if stderr_ is not None:
        stderr = stderr_



import os
import subprocess
import sys
import unittest


class TestOptions(unittest.TestCase):
    def test_clients(self):
        output = run_command(['--clients'])
        expected = get_expected_output('clients')
        self.assertEqual(output, expected)

    def test_commands(self):
        output = run_command(['--commands'])
        expected = get_expected_output('commands')
        self.assertEqual(output, expected)


def run_command(args):
    repo_root = os.path.dirname(os.path.dirname(__file__))
    script = os.path.join(repo_root, 'scripts', 'vcs')
    env = dict(os.environ)
    env.update(
        LANG='en_US.UTF-8',
        PYTHONPATH=repo_root + os.pathsep + env.get('PYTHONPATH', ''),
    )
    return subprocess.check_output(
        [sys.executable, script] + (args or []), stderr=subprocess.STDOUT, env=env
    )


def get_expected_output(name):
    path = os.path.join(os.path.dirname(__file__), name + '.txt')
    with open(path, 'rb') as h:
        return h.read()


if __name__ == '__main__':
    unittest.main()



import argparse
import sys

from vcs2l.commands.command import Command, simple_main
from vcs2l.streams import set_streams


class DiffCommand(Command):
    command = 'diff'
    help = 'Show changes in the working tree'

    def __init__(self, args):
        super(DiffCommand, self).__init__(args)
        self.context = args.context


def get_parser():
    parser = argparse.ArgumentParser(
        description='Show changes in the working tree', prog='vcs diff'
    )
    group = parser.add_argument_group('"diff" command parameters')
    group.add_argument(
        '--context',
        metavar='N',
        type=int,
        help='Generate diffs with <n> lines of context',
    )
    return parser


def main(args=None, stdout=None, stderr=None):
    set_streams(stdout=stdout, stderr=stderr)
    parser = get_parser()
    return simple_main(parser, DiffCommand, args)


if __name__ == '__main__':
    sys.exit(main())



import argparse
import sys

from vcs2l.commands.command import Command, simple_main
from vcs2l.streams import set_streams


class PushCommand(Command):
    command = 'push'
    help = 'Push changes from the working copy to the repository'

    def __init__(self, args):
        super(PushCommand, self).__init__(args)


def get_parser():
    parser = argparse.ArgumentParser(
        description='Push changes from the working copy to the repository',
        prog='vcs push',
    )
    parser.add_argument_group('"push" command parameters')
    return parser


def main(args=None, stdout=None, stderr=None):
    set_streams(stdout=stdout, stderr=stderr)
    parser = get_parser()
    return simple_main(parser, PushCommand, args)


if __name__ == '__main__':
    sys.exit(main())



vcs2l_clients = []

try:
    from .bzr import BzrClient

    vcs2l_clients.append(BzrClient)
except ImportError:
    pass

try:
    from .git import GitClient

    vcs2l_clients.append(GitClient)
except ImportError:
    pass

try:
    from .hg import HgClient

    vcs2l_clients.append(HgClient)
except ImportError:
    pass

try:
    from .svn import SvnClient

    vcs2l_clients.append(SvnClient)
except ImportError:
    pass

try:
    from .tar import TarClient

    vcs2l_clients.append(TarClient)
except ImportError:
    pass

try:
    from .zip import ZipClient

    vcs2l_clients.append(ZipClient)
except ImportError:
    pass

_client_types = [c.type for c in vcs2l_clients]
if len(_client_types) != len(set(_client_types)):
    raise RuntimeError(
        'Multiple vcs clients share the same type: ' + ', '.join(sorted(_client_types))
    )



import logging
import os
import sys
import threading
import traceback
from queue import Empty, Queue

logger = logging.getLogger(__name__)
logging.basicConfig()


# Detect special Windows shells that do not support mixes of
# backslashes & forward slashes; for those shells, we want to
# output POSIX paths, i.e. forward slashes only
windows_force_posix = sys.platform == 'win32' and '/' in os.environ.get('_', '')


def fix_output_path(path):
    return path.replace('\\', '/') if windows_force_posix else path


def output_repositories(clients):
    from vcs2l.streams import stdout

    ordered_clients = {client.path: client for client in clients}
    for k in sorted(ordered_clients.keys()):
        client = ordered_clients[k]
        print('%s (%s)' % (fix_output_path(k), client.__class__.type), file=stdout)


def generate_jobs(clients, command):
    jobs = []
    realpaths = {}
    for client in clients:
        # check if client is a duplicate of another path
        realpath = os.path.realpath(client.path)
        if realpath not in realpaths:
            realpaths[realpath] = [client.path]
        else:
            # override command on client to ignore multiple invocations
            # on same repository
            duplicate_path = realpaths[realpath][0]
            realpaths[realpath].append(client.path)
            method_name = command.__class__.command
            method = getattr(client, method_name, None)
            if method is not None:
                setattr(
                    client, method_name, DuplicateCommandHandler(client, duplicate_path)
                )

        job = {'client': client, 'command': command}
        jobs.append(job)
    return jobs


class DuplicateCommandHandler(object):
    def __init__(self, client, duplicate_path):
        self.client = client
        self.duplicate_path = duplicate_path

    def __call__(self, _command):
        return {
            'cmd': '',
            'cwd': self.client.path,
            'output': "Same repository as '%s'" % self.duplicate_path,
            'returncode': None,
        }


def get_ready_job(jobs):
    for job in jobs:
        if not job.get('depends', set()):
            jobs.remove(job)
            return job
    return None


def execute_jobs(jobs, show_progress=False, number_of_workers=10, debug_jobs=False):
    from vcs2l.streams import stdout

    if debug_jobs:
        logger.setLevel(logging.DEBUG)

    if windows_force_posix:
        logger.debug('force POSIX paths on Windows')

    results = []

    job_queue = Queue()
    result_queue = Queue()

    # create worker threads
    workers = []
    for _ in range(min(number_of_workers, len(jobs))):
        worker = Worker(job_queue, result_queue)
        workers.append(worker)

    # fill job_queue with jobs for each worker
    pending_jobs = list(jobs)
    running_job_paths = []
    while job_queue.qsize() < len(workers):
        job = get_ready_job(pending_jobs)
        if not job:
            break
        running_job_paths.append(job['client'].path)
        logger.debug("started '%s'" % job['client'].path)
        job_queue.put(job)
    logger.debug('ongoing %s' % running_job_paths)

    # start all workers
    [w.start() for w in workers]

    # collect results
    while len(results) < len(jobs):
        (job, result) = result_queue.get()
        logger.debug("finished '%s'" % job['client'].path)
        running_job_paths.remove(result['job']['client'].path)
        if show_progress and len(jobs) > 1:
            if result['returncode'] == NotImplemented:
                stdout.write('s')
            elif result['returncode']:
                stdout.write('E')
            else:
                stdout.write('.')
            if debug_jobs:
                stdout.write('\n')
            stdout.flush()
        result.update(job)
        results.append(result)
        if pending_jobs:
            for pending_job in pending_jobs:
                pending_job.get('depends', set()).discard(job['client'].path)
            while job_queue.qsize() < len(workers):
                job = get_ready_job(pending_jobs)
                if not job:
                    break
                running_job_paths.append(job['client'].path)
                logger.debug("started '%s'" % job['client'].path)
                job_queue.put(job)
            assert running_job_paths
        if running_job_paths:
            logger.debug('ongoing ' + str(running_job_paths))
    if show_progress and len(jobs) > 1 and not debug_jobs:
        print('', file=stdout)  # finish progress line

    # join all workers
    for w in workers:
        w.done = True
    [w.join() for w in workers]
    return results


class Worker(threading.Thread):
    def __init__(self, job_queue, result_queue):
        super(Worker, self).__init__()
        self.daemon = True
        self.done = False
        self.job_queue = job_queue
        self.result_queue = result_queue

    def run(self):
        # process all incoming jobs
        while not self.done:
            try:
                # fetch next job
                job = self.job_queue.get(timeout=0.1)
                # process job
                result = self.process_job(job)
                # send result
                self.result_queue.put((job, result))
            except Empty:
                pass

    def process_job(self, job):
        command = job['command']
        if not command:
            return {'cmd': '', 'job': job, 'output': job['output'], 'returncode': 1}
        method_name = command.__class__.command
        try:
            method = getattr(job['client'], method_name, None)
            if method is None:
                return {
                    'cmd': '%s.%s(%s)'
                    % (
                        job['client'].__class__.type,
                        method_name,
                        job['command'].__class__.command,
                    ),
                    'job': job,
                    'output': "Command '%s' not implemented for client '%s'"
                    % (job['command'].__class__.command, job['client'].__class__.type),
                    'returncode': NotImplemented,
                }
            result = method(job['command'])
            result['job'] = job
            return result
        except Exception as e:
            exc_tb = sys.exc_info()[2]
            filename, lineno, _, _ = traceback.extract_tb(exc_tb)[-1]
            return {
                'cmd': '%s.%s(%s)'
                % (
                    job['client'].__class__.type,
                    method_name,
                    job['command'].__class__.command,
                ),
                'job': job,
                'output': "Invocation of command '%s' on client '%s' failed: "
                '%s: %s (%s:%s)'
                % (
                    job['command'].__class__.command,
                    job['client'].__class__.type,
                    type(e).__name__,
                    e,
                    filename,
                    lineno,
                ),
                'returncode': 1,
            }


def output_result(result, hide_empty=False):
    from vcs2l.streams import stdout

    output = result['output']
    if hide_empty and result['returncode'] is None:
        output = ''
    if result['returncode'] == NotImplemented:
        if output:
            output = ansi('yellowf') + output + ansi('reset')
    elif result['returncode']:
        if not output:
            output = 'Failed with return code %d' % result['returncode']
        output = ansi('redf') + output + ansi('reset')
    elif not result['cmd']:
        if output:
            output = ansi('yellowf') + output + ansi('reset')
    if output or not hide_empty:
        client = result['client']
        print(
            ansi('bluef')
            + '=== '
            + ansi('boldon')
            + fix_output_path(client.path)
            + ansi('boldoff')
            + ' ('
            + client.__class__.type
            + ') ==='
            + ansi('reset'),
            file=stdout,
        )
    if output:
        try:
            print(output, file=stdout)
        except UnicodeEncodeError:
            print(output.encode(sys.getdefaultencoding(), 'replace'), file=stdout)


def output_results(results, output_handler=output_result, hide_empty=False):
    # output results in alphabetic order
    path_to_idx = {result['client'].path: i for i, result in enumerate(results)}
    idxs_in_order = [path_to_idx[path] for path in sorted(path_to_idx.keys())]
    for i in idxs_in_order:
        output_handler(results[i], hide_empty=hide_empty)


USE_COLOR = hasattr(sys.stdout, 'isatty') and sys.stdout.isatty()
# disable color on Windows except if ConEmuANSI is explicitly enabled
if os.name == 'nt' and os.environ.get('ConEmuANSI', None) != 'ON':
    USE_COLOR = False


def ansi(keyword):
    if not USE_COLOR:
        return ''
    codes = {
        'bluef': '\033[34m',
        'boldon': '\033[1m',
        'boldoff': '\033[22m',
        'cyanf': '\033[36m',
        'redf': '\033[31m',
        'reset': '\033[0m',
        'yellowf': '\033[33m',
    }
    if keyword in codes:
        return codes[keyword]
    return ''



import os
import stat
import sys
from errno import EACCES, EPERM
from shutil import rmtree as shutil_rmtree


def rmtree(path):
    kwargs = {}
    if sys.platform == 'win32':
        kwargs['onerror'] = _onerror_windows
    return shutil_rmtree(path, **kwargs)


def _onerror_windows(function, path, excinfo):
    if isinstance(excinfo[1], OSError) and excinfo[1].errno in (EACCES, EPERM):
        os.chmod(path, stat.S_IWRITE)
        function(path)



import os
import zipfile
from io import BytesIO
from urllib.error import URLError

from vcs2l.clients.vcs_base import VcsClientBase, load_url, test_url
from vcs2l.util import rmtree


class ZipClient(VcsClientBase):
    type = 'zip'

    @staticmethod
    def is_repository(path):
        return False

    def __init__(self, path):
        super(ZipClient, self).__init__(path)

    def import_(self, command):
        if not command.url:
            return {
                'cmd': '',
                'cwd': self.path,
                'output': "Repository data lacks the 'url' value",
                'returncode': 1,
            }

        # clear destination
        if os.path.exists(self.path):
            for filename in os.listdir(self.path):
                path = os.path.join(self.path, filename)
                try:
                    rmtree(path)
                except OSError:
                    os.remove(path)
        else:
            not_exist = self._create_path()
            if not_exist:
                return not_exist

        # download zipfile
        try:
            data = load_url(command.url, retry=command.retry)
        except URLError as e:
            return {
                'cmd': '',
                'cwd': self.path,
                'output': "Could not fetch zipfile from '%s': %s" % (command.url, e),
                'returncode': 1,
            }

        def create_path(path):
            if not os.path.exists(path):
                try:
                    os.makedirs(path)
                except os.error as e:
                    return {
                        'cmd': 'os.makedirs(%s)' % path,
                        'cwd': path,
                        'output': "Could not create directory '%s': %s" % (path, e),
                        'returncode': 1,
                    }
            return None

        # unpack zipfile into destination
        try:
            zip_file = zipfile.ZipFile(BytesIO(data), mode='r')
        except zipfile.BadZipfile as e:
            return {
                'cmd': 'ZipFile(%s)' % command.url,
                'cwd': self.path,
                'output': "Could not read zipfile from '%s': %s" % (command.url, e),
                'returncode': 1,
            }
        try:
            if not command.version:
                zip_file.extractall(self.path)
            else:
                prefix = str(command.version) + '/'
                for name in zip_file.namelist():
                    if name.startswith(prefix):
                        if not name[len(prefix) :]:
                            continue
                        # remap members from version subfolder into destination
                        dst = os.path.join(self.path, name[len(prefix) :])
                        if dst.endswith('/'):
                            # create directories
                            not_exist = create_path(dst)
                            if not_exist:
                                return not_exist
                        else:
                            with zip_file.open(name, mode='r') as src_handle:
                                with open(dst, 'wb') as dst_handle:
                                    dst_handle.write(src_handle.read())
        finally:
            zip_file.close()

        return {
            'cmd': '',
            'cwd': self.path,
            'output': "Downloaded zipfile from '%s' and unpacked it" % command.url,
            'returncode': 0,
        }

    def validate(self, command):
        if not command.url:
            return {
                'cmd': '',
                'cwd': self.path,
                'output': "Repository data lacks the 'url' value",
                'returncode': 1,
            }

        # test url
        try:
            test_url(command.url, retry=command.retry)
        except URLError as e:
            return {
                'cmd': '',
                'cwd': self.path,
                'output': "Failed to contact zip url '%s': %s" % (command.url, e),
                'returncode': 1,
            }
        return {
            'cmd': 'http HEAD',
            'cwd': self.path,
            'output': "Zip url '%s' exists" % command.url,
            'returncode': None,
        }



import argparse
import sys

from vcs2l.clients import vcs2l_clients
from vcs2l.commands import vcs2l_commands
from vcs2l.errors import UnsupportedPythonVersionError
from vcs2l.streams import set_streams

if sys.version_info >= (3, 8):
    from importlib.metadata import entry_points
elif sys.version_info >= (3, 7):
    from importlib_metadata import entry_points
elif sys.version_info >= (3, 5):
    from pkg_resources import load_entry_point
else:
    raise UnsupportedPythonVersionError()


def main(args=None, stdout=None, stderr=None):
    set_streams(stdout=stdout, stderr=stderr)

    # no help to extract command first (which might be followed by --help)
    parser = get_parser(add_help=False)
    ns, _ = parser.parse_known_args(args)

    # help for a specific command
    if ns.command:
        # relay help request foe specific command
        entrypoint = get_entrypoint(ns.command)
        if not entrypoint:
            return 1
        return entrypoint(['--help'])

    # regular parsing validating options and arguments
    parser = get_parser()
    ns = parser.parse_args(args)

    if ns.clients:
        print('The available VCS clients are:')
        for client in vcs2l_clients:
            print('  ' + client.type)
        return 0

    if ns.commands:
        print(' '.join([cmd.command for cmd in vcs2l_commands]))
        return 0

    if ns.commands_descriptions:
        print(
            '\n'.join(
                ['{}\t{}'.format(cmd.command, cmd.help) for cmd in vcs2l_commands]
            )
        )
        return 0

    # output detailed command list
    parser = get_parser_with_command_only()
    parser.print_help()
    return 0


def get_parser(add_help=True):
    parser = argparse.ArgumentParser(
        prog='vcs',
        description=_get_description(),
        epilog=_get_epilog(),
        add_help=add_help,
    )
    group = parser.add_mutually_exclusive_group()
    group.add_argument(
        'command',
        metavar='<command>',
        nargs='?',
        help='The available commands: '
        + ', '.join([cmd.command for cmd in vcs2l_commands]),
    )
    group.add_argument(
        '--clients',
        action='store_true',
        default=False,
        help='Show the available VCS clients',
    )
    group.add_argument(
        '--commands',
        action='store_true',
        default=False,
        help='Output the available commands for auto-completion',
    )
    group.add_argument(
        '--commands-descriptions',
        action='store_true',
        default=False,
        help='Output the available commands along with their descriptions',
    )
    from vcs2l import __version__

    group.add_argument(
        '--version',
        action='version',
        version='%(prog)s ' + __version__,
        help='Show the vcs2l version',
    )
    return parser


def get_entrypoint(command):
    # accept command with same prefix if unique
    commands = [cmd.command for cmd in vcs2l_commands]
    commands = [cmd for cmd in commands if cmd.startswith(command)]
    if len(commands) != 1:
        print(
            "vcs: '%s' is not a vcs command. See 'vcs help'." % command, file=sys.stderr
        )
        if commands:
            print(
                '\nDid you mean one of these?\n' + '\n   '.join(commands),
                file=sys.stderr,
            )
        return None

    ep_name = 'vcs-' + commands[0]

    if sys.version_info >= (3, 10):
        eps = entry_points()
        entry_point = next(iter(eps.select(group='console_scripts', name=ep_name)))
        if entry_point:
            return entry_point.load()

    elif sys.version_info >= (3, 7):
        eps = entry_points()
        for ep in eps.get('console_scripts', []):
            if ep.name == ep_name:
                return ep.load()

    elif sys.version_info >= (3, 5):
        return load_entry_point('vcs2l', 'console_scripts', ep_name)

    else:
        raise UnsupportedPythonVersionError()

    return None


def get_parser_with_command_only():
    parser = argparse.ArgumentParser(
        prog='vcs',
        usage='%(prog)s <command>',
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description='%s\n\n%s'
        % (_get_description(), '\n'.join(_get_command_help(vcs2l_commands))),
        epilog=_get_epilog(),
        add_help=False,
    )
    parser.add_argument('command', help=argparse.SUPPRESS)
    return parser


def _get_description():
    return (
        'Most commands take directory arguments, '
        'recursively searching for repositories\n'
        'in these directories.  '
        'If no arguments are supplied to a command, it recurses\n'
        'on the current directory (inclusive) by default.'
    )


def _get_epilog():
    return "See '%(prog)s <command> --help' for more information on a specific command."


def _get_command_help(commands):
    lines = ['The available commands are:']
    max_len = max(len(cmd.command) for cmd in commands)
    for cmd in vcs2l_commands:
        lines.append(
            '   %s%s   %s' % (cmd.command, ' ' * (max_len - len(cmd.command)), cmd.help)
        )
    return lines


if __name__ == '__main__':
    sys.exit(main())



import os
from shutil import which
from threading import Lock

from vcs2l.clients.vcs_base import VcsClientBase
from vcs2l.executor import USE_COLOR
from vcs2l.util import rmtree


class HgClient(VcsClientBase):
    type = 'hg'
    _executable = None
    _config_color = None
    _config_color_lock = Lock()

    @staticmethod
    def is_repository(path):
        return os.path.isdir(os.path.join(path, '.hg'))

    def __init__(self, path):
        super(HgClient, self).__init__(path)

    def branch(self, command):
        self._check_executable()
        cmd = [HgClient._executable, 'branches' if command.all else 'branch']
        self._check_color(cmd)
        return self._run_command(cmd)

    def custom(self, command):
        self._check_executable()
        cmd = [HgClient._executable] + command.args
        return self._run_command(cmd)

    def diff(self, command):
        self._check_executable()
        cmd = [HgClient._executable, 'diff']
        self._check_color(cmd)
        if command.context:
            cmd += ['--unified %d' % command.context]
        return self._run_command(cmd)

    def export(self, command):
        self._check_executable()
        result_url = self._get_url()
        if result_url['returncode']:
            return result_url
        url = result_url['output']

        cmd_id = [HgClient._executable, 'identify', '--id']
        result_id = self._run_command(cmd_id)
        if result_id['returncode']:
            result_id['output'] = 'Could not determine id: ' + result_id['output']
            return result_id
        id_ = result_id['output']

        if not command.exact:
            cmd_branch = [HgClient._executable, 'identify', '--branch']
            result_branch = self._run_command(cmd_branch)
            if result_branch['returncode']:
                result_branch['output'] = (
                    'Could not determine branch: ' + result_branch['output']
                )
                return result_branch
            branch = result_branch['output']

            cmd_branch_id = [HgClient._executable, 'identify', '-r', branch, '--id']
            result_branch_id = self._run_command(cmd_branch_id)
            if result_branch_id['returncode']:
                result_branch_id['output'] = (
                    'Could not determine branch id: ' + result_branch_id['output']
                )
                return result_branch_id
            if result_branch_id['output'] == id_:
                id_ = branch
                cmd_branch = cmd_branch_id

        return {
            'cmd': '%s && %s' % (result_url['cmd'], ' '.join(cmd_id)),
            'cwd': self.path,
            'output': '\n'.join([url, id_]),
            'returncode': 0,
            'export_data': {'url': url, 'version': id_},
        }

    def _get_url(self):
        cmd_url = [HgClient._executable, 'paths', 'default']
        result_url = self._run_command(cmd_url)
        if result_url['returncode']:
            result_url['output'] = 'Could not determine url: ' + result_url['output']
            return result_url
        return result_url

    def import_(self, command):
        if not command.url or not command.version:
            if not command.url and not command.version:
                value_missing = "'url' and 'version'"
            elif not command.url:
                value_missing = "'url'"
            else:
                value_missing = "'version'"
            return {
                'cmd': '',
                'cwd': self.path,
                'output': 'Repository data lacks the %s value' % value_missing,
                'returncode': 1,
            }

        self._check_executable()
        if HgClient.is_repository(self.path):
            # verify that existing repository is the same
            result_url = self._get_url()
            if result_url['returncode']:
                return result_url
            url = result_url['output']
            if url != command.url:
                if not command.force:
                    return {
                        'cmd': '',
                        'cwd': self.path,
                        'output': 'Path already exists and contains a different '
                        'repository',
                        'returncode': 1,
                    }
                try:
                    rmtree(self.path)
                except OSError:
                    os.remove(self.path)

        not_exist = self._create_path()
        if not_exist:
            return not_exist

        if HgClient.is_repository(self.path):
            # pull updates for existing repo
            cmd_pull = [HgClient._executable, '--noninteractive', 'pull', '--update']
            result_pull = self._run_command(cmd_pull, retry=command.retry)
            if result_pull['returncode']:
                return result_pull
            cmd = result_pull['cmd']
            output = result_pull['output']

        else:
            cmd_clone = [
                HgClient._executable,
                '--noninteractive',
                'clone',
                command.url,
                '.',
            ]
            result_clone = self._run_command(cmd_clone, retry=command.retry)
            if result_clone['returncode']:
                result_clone['output'] = "Could not clone repository '%s': %s" % (
                    command.url,
                    result_clone['output'],
                )
                return result_clone
            cmd = result_clone['cmd']
            output = result_clone['output']

        if command.version:
            cmd_checkout = [
                HgClient._executable,
                '--noninteractive',
                'checkout',
                command.version,
            ]
            result_checkout = self._run_command(cmd_checkout)
            if result_checkout['returncode']:
                result_checkout['output'] = "Could not checkout '%s': %s" % (
                    command.version,
                    result_checkout['output'],
                )
                return result_checkout
            cmd += ' && ' + ' '.join(cmd_checkout)
            output = '\n'.join([output, result_checkout['output']])

        return {'cmd': cmd, 'cwd': self.path, 'output': output, 'returncode': 0}

    def log(self, command):
        self._check_executable()
        if command.limit_tag:
            # check if specific tag exists
            cmd_log = [
                HgClient._executable,
                'log',
                '--rev',
                'tag(%s)' % command.limit_tag,
            ]
            result_log = self._run_command(cmd_log)
            if result_log['returncode']:
                return {
                    'cmd': '',
                    'cwd': self.path,
                    'output': "Repository lacks the tag '%s'" % command.limit_tag,
                    'returncode': 1,
                }
            # output log since specific tag
            cmd = [
                HgClient._executable,
                'log',
                '--rev',
                'sort(tag(%s)::, -rev) and not tag (%s)'
                % (command.limit_tag, command.limit_tag),
            ]
        elif command.limit_untagged:
            # determine distance to nearest tag
            cmd_log = [
                HgClient._executable,
                'log',
                '--rev',
                '.',
                '--template',
                '{latesttagdistance}',
            ]
            result_log = self._run_command(cmd_log)
            if result_log['returncode']:
                return result_log
            # output log since nearest tag
            cmd = [
                HgClient._executable,
                'log',
                '--limit',
                result_log['output'],
                '-b',
                '.',
            ]
        else:
            cmd = [HgClient._executable, 'log']
        if command.limit != 0:
            cmd += ['--limit', '%d' % command.limit]
        if command.verbose:
            cmd += ['--verbose']
        self._check_color(cmd)
        return self._run_command(cmd)

    def pull(self, _command):
        self._check_executable()
        cmd = [HgClient._executable, '--noninteractive', 'pull', '--update']
        self._check_color(cmd)
        return self._run_command(cmd)

    def push(self, _command):
        self._check_executable()
        cmd = [HgClient._executable, '--noninteractive', 'push']
        return self._run_command(cmd)

    def remotes(self, _command):
        self._check_executable()
        cmd = [HgClient._executable, 'paths']
        return self._run_command(cmd)

    def status(self, command):
        self._check_executable()
        cmd = [HgClient._executable, 'status']
        self._check_color(cmd)
        if command.quiet:
            cmd += ['--untracked-files=no']
        return self._run_command(cmd)

    def validate(self, command):
        if not command.url:
            return {
                'cmd': '',
                'cwd': self.path,
                'output': "Repository data lacks the 'url' value",
                'returncode': 1,
            }

        self._check_executable()

        cmd_id_repo = [
            HgClient._executable,
            '--noninteractive',
            'identify',
            command.url,
        ]
        result_id_repo = self._run_command(cmd_id_repo, retry=command.retry)
        if result_id_repo['returncode']:
            result_id_repo['output'] = (
                "Failed to contact remote repository '%s': %s"
                % (command.url, result_id_repo['output'])
            )
            return result_id_repo

        if command.version:
            cmd_id_ver = [
                HgClient._executable,
                '--noninteractive',
                'identify',
                '-r',
                command.version,
                command.url,
            ]
            result_id_ver = self._run_command(cmd_id_ver, retry=command.retry)
            if result_id_ver['returncode']:
                result_id_ver['output'] = (
                    'Specified version not found on remote repository '
                    + "'%s':'%s' : %s"
                    % (command.url, command.version, result_id_ver['output'])
                )
                return result_id_ver

            cmd = result_id_ver['cmd']
            output = "Found hg repository '%s' with changeset '%s'" % (
                command.url,
                command.version,
            )
        else:
            cmd = result_id_repo['cmd']
            output = "Found hg repository '%s' with default branch" % command.url

        return {'cmd': cmd, 'cwd': self.path, 'output': output, 'returncode': None}

    def _check_color(self, cmd):
        if not USE_COLOR:
            return
        with HgClient._config_color_lock:
            # check if user uses colorization
            if HgClient._config_color is None:
                HgClient._config_color = False
                # check if config extension is available
                _cmd = [HgClient._executable, 'config', '--help']
                result = self._run_command(_cmd)
                if result['returncode']:
                    return
                # check if color extension is available and not disabled
                _cmd = [HgClient._executable, 'config', 'extensions.color']
                result = self._run_command(_cmd)
                if result['returncode'] or result['output'].startswith('!'):
                    return
                # check if color mode is not off or not set
                _cmd = [HgClient._executable, 'config', 'color.mode']
                result = self._run_command(_cmd)
                if not result['returncode'] and result['output'] == 'off':
                    return
                HgClient._config_color = True

        # inject arguments to force colorization
        if HgClient._config_color:
            cmd[1:1] = '--color', 'always'

    def _check_executable(self):
        assert HgClient._executable is not None, "Could not find 'hg' executable"


if not HgClient._executable:
    HgClient._executable = which('hg')



import os
import subprocess
from shutil import which

from vcs2l.clients.vcs_base import VcsClientBase
from vcs2l.executor import USE_COLOR
from vcs2l.util import rmtree


class GitClient(VcsClientBase):
    type = 'git'
    _executable = None
    _git_version = None
    _config_color_is_auto = None

    @classmethod
    def get_git_version(cls):
        if cls._git_version is None:
            output = subprocess.check_output(['git', '--version'])
            prefix = b'git version '
            assert output.startswith(prefix)
            output = output[len(prefix) :].split(maxsplit=1)[0]
            cls._git_version = [int(x) for x in output.split(b'.') if x != b'windows']
        return cls._git_version

    @staticmethod
    def is_repository(path):
        return os.path.isdir(os.path.join(path, '.git'))

    def __init__(self, path):
        super(GitClient, self).__init__(path)

    def branch(self, command):
        self._check_executable()
        cmd = [GitClient._executable, 'branch']
        result = self._run_command(cmd)

        if not command.all and not result['returncode']:
            # only show current branch
            lines = result['output'].splitlines()
            lines = [line[2:] for line in lines if line.startswith('* ')]
            result['output'] = '\n'.join(lines)

        return result

    def custom(self, command):
        self._check_executable()
        cmd = [GitClient._executable] + command.args
        return self._run_command(cmd)

    def diff(self, command):
        self._check_executable()
        cmd = [GitClient._executable, 'diff']
        self._check_color(cmd)
        if command.context:
            cmd += ['--unified=%d' % command.context]
        return self._run_command(cmd)

    def export(self, command):
        self._check_executable()
        exact = command.exact
        if not exact:
            # determine if a specific branch is checked out or ec is detached
            cmd_branch = [GitClient._executable, 'rev-parse', '--abbrev-ref', 'HEAD']
            result_branch = self._run_command(cmd_branch)
            if result_branch['returncode']:
                result_branch['output'] = (
                    'Could not determine ref: ' + result_branch['output']
                )
                return result_branch
            branch_name = result_branch['output']
            exact = branch_name == 'HEAD'  # is detached

        if not exact:
            # determine the remote of the current branch
            cmd_remote = [
                GitClient._executable,
                'rev-parse',
                '--abbrev-ref',
                '@{upstream}',
            ]
            result_remote = self._run_command(cmd_remote)
            if result_remote['returncode']:
                result_remote['output'] = (
                    'Could not determine ref: ' + result_remote['output']
                )
                return result_remote
            branch_with_remote = result_remote['output']

            # determine remote
            suffix = '/' + branch_name
            assert branch_with_remote.endswith(branch_name), (
                "'%s' does not end with '%s'" % (branch_with_remote, branch_name)
            )
            remote = branch_with_remote[: -len(suffix)]

            # if a local ref exists with the same name as the remote branch
            # the result will be prefixed to make it unambiguous
            prefix = 'remotes/'
            if remote.startswith(prefix):
                remote = remote[len(prefix) :]

            # determine url of remote
            result_url = self._get_remote_url(remote)
            if result_url['returncode']:
                return result_url
            url = result_url['output']

            # the result is the remote url and the branch name
            return {
                'cmd': ' && '.join(
                    [result_branch['cmd'], result_remote['cmd'], result_url['cmd']]
                ),
                'cwd': self.path,
                'output': '\n'.join([url, branch_name]),
                'returncode': 0,
                'export_data': {'url': url, 'version': branch_name},
            }

        else:
            # determine the hash
            cmd_ref = [GitClient._executable, 'rev-parse', 'HEAD']
            result_ref = self._run_command(cmd_ref)
            if result_ref['returncode']:
                result_ref['output'] = (
                    'Could not determine ref: ' + result_ref['output']
                )
                return result_ref
            ref = result_ref['output']

            # get all remote names
            cmd_remotes = [GitClient._executable, 'remote']
            result_remotes = self._run_command(cmd_remotes)
            if result_remotes['returncode']:
                result_remotes['output'] = (
                    'Could not determine remotes: ' + result_remotes['output']
                )
                return result_remotes
            remotes = result_remotes['output'].splitlines()

            # prefer origin and upstream remotes
            if 'upstream' in remotes:
                remotes.remove('upstream')
                remotes.insert(0, 'upstream')
            if 'origin' in remotes:
                remotes.remove('origin')
                remotes.insert(0, 'origin')

            # for each remote name check if the hash is part of the remote
            for remote in remotes:
                # get all remote names
                cmd_refs = [
                    GitClient._executable,
                    'rev-list',
                    '--remotes=' + remote,
                    '--tags',
                ]
                result_refs = self._run_command(cmd_refs)
                if result_refs['returncode']:
                    result_refs['output'] = (
                        "Could not determine refs of remote '%s': " % remote
                        + result_refs['output']
                    )
                    return result_refs
                refs = result_refs['output'].splitlines()
                if ref not in refs:
                    continue

                cmds = [result_ref['cmd']]
                if command.with_tags:
                    # check if there is exactly one tag pointing to that ref
                    cmd_tags = [GitClient._executable, 'tag', '--points-at', ref]
                    result_tags = self._run_command(cmd_tags)
                    if result_tags['returncode']:
                        result_tags['output'] = (
                            "Could not determine tags for ref '%s': " % ref
                            + result_tags['output']
                        )
                        return result_tags
                    cmds.append(result_tags['cmd'])
                    tags = result_tags['output'].splitlines()
                    if len(tags) == 1:
                        tag = tags[0]
                        # double check that the tag is part of the remote
                        # and references the same hash
                        cmd_ls_remote = [
                            GitClient._executable,
                            'ls-remote',
                            remote,
                            'refs/tags/' + tag,
                        ]
                        result_ls_remote = self._run_command(cmd_ls_remote)
                        if result_ls_remote['returncode']:
                            result_ls_remote['output'] = (
                                "Could not check remote tags for '%s': " % remote
                                + result_ls_remote['output']
                            )
                            return result_ls_remote
                        matches = self._get_hash_ref_tuples(result_ls_remote['output'])
                        if len(matches) == 1 and matches[0][0] == ref:
                            ref = tag

                # determine url of remote
                result_url = self._get_remote_url(remote)
                if result_url['returncode']:
                    return result_url
                url = result_url['output']
                cmds.append(result_url['cmd'])

                # the result is the remote url and the hash/tag
                return {
                    'cmd': ' && '.join(cmds),
                    'cwd': self.path,
                    'output': '\n'.join([url, ref]),
                    'returncode': 0,
                    'export_data': {'url': url, 'version': ref},
                }

            return {
                'cmd': ' && '.join([result_ref['cmd'], result_remotes['cmd']]),
                'cwd': self.path,
                'output': "Could not determine remote containing '%s'" % ref,
                'returncode': 1,
            }

    def _get_remote_url(self, remote):
        cmd_url = [GitClient._executable, 'config', '--get', 'remote.%s.url' % remote]
        result_url = self._run_command(cmd_url)
        if result_url['returncode']:
            result_url['output'] = (
                'Could not determine remote url: ' + result_url['output']
            )
        return result_url

    def import_(self, command):
        if not command.url:
            return {
                'cmd': '',
                'cwd': self.path,
                'output': "Repository data lacks the 'url' value",
                'returncode': 1,
            }

        self._check_executable()
        if GitClient.is_repository(self.path):
            # verify that existing repository is the same
            result_urls = self._get_remote_urls()
            if result_urls['returncode']:
                return result_urls
            for url, remote in result_urls['output']:
                if url == command.url:
                    break
            else:
                if command.skip_existing:
                    return {
                        'cmd': '',
                        'cwd': self.path,
                        'output': 'Skipped existing repository with different URL',
                        'returncode': 0,
                    }
                if not command.force:
                    return {
                        'cmd': '',
                        'cwd': self.path,
                        'output': 'Path already exists and contains a different '
                        'repository',
                        'returncode': 1,
                    }
                try:
                    rmtree(self.path)
                except OSError:
                    os.remove(self.path)

        elif command.skip_existing and os.path.exists(self.path):
            return {
                'cmd': '',
                'cwd': self.path,
                'output': 'Skipped existing directory',
                'returncode': 0,
            }

        elif command.force and os.path.exists(self.path):
            # Not empty, not a git repository
            try:
                rmtree(self.path)
            except OSError:
                os.remove(self.path)

        not_exist = self._create_path()
        if not_exist:
            return not_exist

        if GitClient.is_repository(self.path):
            if command.skip_existing:
                checkout_version = None
            elif command.version:
                checkout_version = command.version
            else:
                # determine remote HEAD branch
                cmd_remote = [GitClient._executable, 'remote', 'show', remote]
                # override locale in order to parse output
                env = os.environ.copy()
                env['LC_ALL'] = 'C'
                result_remote = self._run_command(cmd_remote, env=env)
                if result_remote['returncode']:
                    result_remote['output'] = (
                        'Could not get remote information of repository '
                        "'%s': %s" % (url, result_remote['output'])
                    )
                    return result_remote
                prefix = '  HEAD branch: '
                for line in result_remote['output'].splitlines():
                    if line.startswith(prefix):
                        checkout_version = line[len(prefix) :]
                        break
                else:
                    result_remote['returncode'] = 1
                    result_remote['output'] = (
                        'Could not determine remote HEAD branch of '
                        "repository '%s': %s" % (url, result_remote['output'])
                    )
                    return result_remote

            # fetch updates for existing repo
            cmd_fetch = [GitClient._executable, 'fetch', remote]
            if command.shallow:
                result_version_type, version_name = self._check_version_type(
                    command.url, checkout_version, command.retry
                )
                if result_version_type['returncode']:
                    return result_version_type
                version_type = result_version_type['version_type']
                if version_type == 'branch':
                    cmd_fetch.append(
                        'refs/heads/%s:refs/remotes/%s/%s'
                        % (version_name, remote, version_name)
                    )
                elif version_type == 'hash':
                    cmd_fetch.append(checkout_version)
                elif version_type == 'tag':
                    cmd_fetch.append(
                        '+refs/tags/%s:refs/tags/%s' % (version_name, version_name)
                    )
                else:
                    assert False
                cmd_fetch += ['--depth', '1']
            else:
                version_type = None
            result_fetch = self._run_command(cmd_fetch, retry=command.retry)
            if result_fetch['returncode']:
                return result_fetch
            cmd = result_fetch['cmd']
            output = result_fetch['output']

            if not command.shallow and checkout_version is not None:
                if checkout_version.startswith('heads/'):
                    version_name = checkout_version[6:]
                    version_type = 'branch'
                elif checkout_version.startswith('tags/'):
                    version_name = checkout_version[5:]
                    version_type = 'tag'
                else:
                    version_type = None
            # ensure that a tracking branch exists which can be checked out
            if version_type == 'branch':
                cmd_show_ref = [
                    GitClient._executable,
                    'show-ref',
                    'refs/heads/%s' % version_name,
                ]
                result_show_ref = self._run_command(cmd_show_ref)
                if result_show_ref['returncode']:
                    if not command.shallow:
                        result_show_ref['output'] = "Could not find branch '%s': %s" % (
                            version_name,
                            result_show_ref['output'],
                        )
                        return result_show_ref
                    # creating tracking branch
                    cmd_branch = [
                        GitClient._executable,
                        'branch',
                        version_name,
                        '%s/%s' % (remote, version_name),
                    ]
                    result_branch = self._run_command(cmd_branch)
                    if result_branch['returncode']:
                        result_branch['output'] = "Could not create branch '%s': %s" % (
                            version_name,
                            result_branch['output'],
                        )
                        return result_branch
                    cmd += ' && ' + ' '.join(cmd_branch)
                    output = '\n'.join([output, result_branch['output']])
                checkout_version = version_name

        else:
            version_type = None
            if command.version:
                result_version_type, version_name = self._check_version_type(
                    command.url, command.version, command.retry
                )
                if result_version_type['returncode']:
                    return result_version_type
                version_type = result_version_type['version_type']

            if not command.shallow or version_type in (None, 'branch'):
                cmd_clone = [GitClient._executable, 'clone', command.url, '.']
                if version_type == 'branch':
                    cmd_clone += ['-b', version_name]
                    checkout_version = None
                else:
                    checkout_version = command.version
                if command.shallow:
                    cmd_clone += ['--depth', '1']
                result_clone = self._run_command(cmd_clone, retry=command.retry)
                if result_clone['returncode']:
                    result_clone['output'] = "Could not clone repository '%s': %s" % (
                        command.url,
                        result_clone['output'],
                    )
                    return result_clone
                cmd = result_clone['cmd']
                output = result_clone['output']
            else:
                # getting a hash or tag with a depth of 1 can't use 'clone'
                cmd_init = [GitClient._executable, 'init']
                result_init = self._run_command(cmd_init)
                if result_init['returncode']:
                    return result_init
                cmd = result_init['cmd']
                output = result_init['output']

                cmd_remote_add = [
                    GitClient._executable,
                    'remote',
                    'add',
                    'origin',
                    command.url,
                ]
                result_remote_add = self._run_command(cmd_remote_add)
                if result_remote_add['returncode']:
                    return result_remote_add
                cmd += ' && ' + ' '.join(cmd_remote_add)
                output = '\n'.join([output, result_remote_add['output']])

                cmd_fetch = [GitClient._executable, 'fetch', 'origin']
                if version_type == 'hash':
                    cmd_fetch.append(command.version)
                elif version_type == 'tag':
                    cmd_fetch.append(
                        'refs/tags/%s:refs/tags/%s' % (version_name, version_name)
                    )
                else:
                    assert False
                cmd_fetch += ['--depth', '1']
                result_fetch = self._run_command(cmd_fetch, retry=command.retry)
                if result_fetch['returncode']:
                    return result_fetch
                cmd += ' && ' + ' '.join(cmd_fetch)
                output = '\n'.join([output, result_fetch['output']])

                checkout_version = command.version

        if checkout_version:
            cmd_checkout = [GitClient._executable, 'checkout', checkout_version, '--']
            result_checkout = self._run_command(cmd_checkout)
            if result_checkout['returncode']:
                if self.get_git_version() < [1, 8, 4, 3]:
                    cmd_checkout.pop()
                    result_checkout = self._run_command(cmd_checkout)
            if result_checkout['returncode']:
                result_checkout['output'] = "Could not checkout ref '%s': %s" % (
                    checkout_version,
                    result_checkout['output'],
                )
                return result_checkout
            cmd += ' && ' + ' '.join(cmd_checkout)
            output = '\n'.join([output, result_checkout['output']])

        if command.recursive:
            cmd_submodule = [
                GitClient._executable,
                'submodule',
                'update',
                '--init',
                '--recursive',
            ]
            result_submodule = self._run_command(cmd_submodule)
            if result_submodule['returncode']:
                result_submodule['output'] = (
                    'Could not init/update submodules: %s' % result_submodule['output']
                )
                return result_submodule
            cmd += ' && ' + ' '.join(cmd_submodule)
            output = '\n'.join([output, result_submodule['output']])

        return {'cmd': cmd, 'cwd': self.path, 'output': output, 'returncode': 0}

    def _get_remote_urls(self):
        cmd_remote = [GitClient._executable, 'remote', 'show']
        result_remote = self._run_command(cmd_remote)
        if result_remote['returncode']:
            result_remote['output'] = (
                'Could not determine remotes: ' + result_remote['output']
            )
            return result_remote
        remote_urls = []
        cmd = result_remote['cmd']
        for remote in result_remote['output'].splitlines():
            result_url = self._get_remote_url(remote)
            cmd += ' && ' + result_url['cmd']
            if not result_url['returncode']:
                remote_urls.append((result_url['output'], remote))
        return {
            'cmd': cmd,
            'cwd': self.path,
            'output': (
                remote_urls
                if remote_urls
                else 'Could not determine any of the remote urls'
            ),
            'returncode': 0 if remote_urls else 1,
        }

    def _check_version_type(self, url, version, retry=0):
        # check if version starts with heads/ or tags/
        prefixes = {
            'heads/': 'branch',
            'tags/': 'tag',
        }
        for prefix, version_type in prefixes.items():
            if version.startswith(prefix):
                return {
                    'cmd': None,
                    'cwd': None,
                    'output': None,
                    'returncode': 0,
                    'version_type': version_type,
                }, version[len(prefix) :]

        cmd = [GitClient._executable, 'ls-remote', url, version]
        result = self._run_command(cmd, retry=retry)
        if result['returncode']:
            result['output'] = (
                'Could not determine ref type of version: ' + result['output']
            )
            return result, None
        if not result['output']:
            result['version_type'] = 'hash'
            return result, None

        refs = {}
        for hash_, ref in self._get_hash_ref_tuples(result['output']):
            refs[ref] = hash_

        tag_ref = 'refs/tags/' + version
        branch_ref = 'refs/heads/' + version
        if tag_ref in refs and branch_ref in refs:
            if refs[tag_ref] != refs[branch_ref]:
                result['returncode'] = 1
                result['output'] = (
                    'The version ref is a branch as well as '
                    'tag but with different hashes'
                )
                return result, None
        if tag_ref in refs:
            result['version_type'] = 'tag'
        elif branch_ref in refs:
            result['version_type'] = 'branch'
        else:
            result['version_type'] = 'hash'
        return result, version if result['version_type'] in ('tag', 'branch') else None

    def log(self, command):
        self._check_executable()
        if command.limit_tag:
            # check if specific tag exists
            cmd_tag = [GitClient._executable, 'tag', '-l', command.limit_tag]
            result_tag = self._run_command(cmd_tag)
            if result_tag['returncode']:
                return result_tag
            if not result_tag['output']:
                return {
                    'cmd': '',
                    'cwd': self.path,
                    'output': "Repository lacks the tag '%s'" % command.limit_tag,
                    'returncode': 1,
                }
            # output log since specific tag
            cmd = [GitClient._executable, 'log', '%s..' % command.limit_tag]
        elif command.limit_untagged:
            # determine nearest tag
            cmd_tag = [GitClient._executable, 'describe', '--abbrev=0', '--tags']
            result_tag = self._run_command(cmd_tag)
            if result_tag['returncode']:
                return result_tag
            # output log since nearest tag
            cmd = [GitClient._executable, 'log', '%s..' % result_tag['output']]
        else:
            cmd = [GitClient._executable, 'log']
        if command.merge_only:
            cmd += ['--merges']
        cmd += ['--decorate']
        if command.limit != 0:
            cmd += ['-%d' % command.limit]
        if not command.verbose:
            cmd += ['--pretty=short']
        self._check_color(cmd)
        return self._run_command(cmd)

    def pull(self, _command):
        self._check_executable()
        cmd = [GitClient._executable, 'pull']
        self._check_color(cmd)
        result = self._run_command(cmd)

        if result['returncode']:
            # check for detached HEAD
            cmd_rev_parse = [GitClient._executable, 'rev-parse', '--abbrev-ref', 'HEAD']
            result_rev_parse = self._run_command(cmd_rev_parse)
            if result_rev_parse['returncode']:
                result_rev_parse['output'] = (
                    'Could not determine ref: ' + result_rev_parse['output']
                )
                return result_rev_parse
            detached = result_rev_parse['output'] == 'HEAD'

            if detached:
                # warn but not fail about the inability to pull a detached head
                return {
                    'cmd': '',
                    'cwd': self.path,
                    'output': result['output'],
                    'returncode': 0,
                }

        return result

    def push(self, _command):
        self._check_executable()
        cmd = [GitClient._executable, 'push']
        return self._run_command(cmd)

    def remotes(self, _command):
        self._check_executable()
        cmd = [GitClient._executable, 'remote', '-v']
        return self._run_command(cmd)

    def status(self, command):
        self._check_executable()
        while command.hide_empty:
            # check if ahead
            cmd = [GitClient._executable, 'log', '@{push}..']
            result = self._run_command(cmd)
            if not result['returncode'] and result['output']:
                # ahead, do not hide
                break
            # check if behind
            cmd = [GitClient._executable, 'log', '..@{upstream}']
            result = self._run_command(cmd)
            if not result['returncode'] and result['output']:
                # behind, do not hide
                break
            cmd = [GitClient._executable, 'status', '-s']
            if command.quiet:
                cmd += ['--untracked-files=no']
            result = self._run_command(cmd)
            if result['returncode'] or not result['output']:
                return result
            break
        cmd = [GitClient._executable, 'status']
        self._check_color(cmd)
        if command.quiet:
            cmd += ['--untracked-files=no']
        return self._run_command(cmd)

    def validate(self, command):
        if not command.url:
            return {
                'cmd': '',
                'cwd': self.path,
                'output': "Repository data lacks the 'url' value",
                'returncode': 1,
            }

        self._check_executable()

        cmd_ls_remote = [GitClient._executable, 'ls-remote']
        cmd_ls_remote += ['-q', '--exit-code']
        cmd_ls_remote += [command.url]
        env = os.environ.copy()
        env['GIT_TERMINAL_PROMPT'] = '0'
        result_ls_remote = self._run_command(
            cmd_ls_remote, retry=command.retry, env=env
        )
        if result_ls_remote['returncode']:
            result_ls_remote['output'] = (
                "Failed to contact remote repository '%s': %s"
                % (command.url, result_ls_remote['output'])
            )
            return result_ls_remote

        if command.version:
            hashes = []
            refs = []
            tags = []
            branches = []

            for hash_and_ref in self._get_hash_ref_tuples(result_ls_remote['output']):
                hashes.append(hash_and_ref[0])

                # ignore pull request refs
                if not hash_and_ref[1].startswith('refs/pull/'):
                    if hash_and_ref[1].startswith('refs/tags/'):
                        tags.append(hash_and_ref[1][10:])
                    elif hash_and_ref[1].startswith('refs/heads/'):
                        branches.append(hash_and_ref[1][11:])
                    else:
                        refs.append(hash_and_ref[1])

            if command.version in refs:
                version_type = 'ref'
                version_name = command.version
            elif (
                command.version.startswith('heads/') and command.version[6:] in branches
            ):
                version_type = 'branch'
                version_name = command.version[6:]
            elif command.version.startswith('tags/') and command.version[5:] in tags:
                version_type = 'tag'
                version_name = command.version[5:]
            elif command.version in branches and command.version not in tags:
                version_type = 'branch'
                version_name = command.version
            elif command.version in tags and command.version not in branches:
                version_type = 'tag'
                version_name = command.version
            else:
                for _hash in hashes:
                    if _hash.startswith(command.version):
                        version_type = 'hash'
                        version_name = command.version
                        break
                else:
                    cmd = result_ls_remote['cmd']
                    output = (
                        "Found git repository '%s' but " % command.url
                        + 'unable to verify non-branch / non-tag ref '
                        + "'%s' without cloning the repo" % command.version
                    )

                    return {
                        'cmd': cmd,
                        'cwd': self.path,
                        'output': output,
                        'returncode': 0,
                    }

            cmd = result_ls_remote['cmd']
            output = "Found git repository '%s' with %s '%s'" % (
                command.url,
                version_type,
                version_name,
            )
        else:
            cmd = result_ls_remote['cmd']
            output = "Found git repository '%s' with default branch" % command.url

        return {'cmd': cmd, 'cwd': self.path, 'output': output, 'returncode': None}

    def _check_color(self, cmd):
        if not USE_COLOR:
            return
        # check if user uses colorization
        if GitClient._config_color_is_auto is None:
            _cmd = [GitClient._executable, 'config', '--get', 'color.ui']
            result = self._run_command(_cmd)
            GitClient._config_color_is_auto = result['output'] in ['', 'auto']

        # inject arguments to force colorization
        if GitClient._config_color_is_auto:
            cmd[1:1] = '-c', 'color.ui=always'

    def _check_executable(self):
        assert GitClient._executable is not None, "Could not find 'git' executable"

    def _get_hash_ref_tuples(self, ls_remote_output):
        tuples = []
        for line in ls_remote_output.splitlines():
            if line.startswith('#'):
                continue
            try:
                hash_, ref = line.split(None, 1)
            except ValueError:
                continue
            tuples.append((hash_, ref))
        return tuples


if not GitClient._executable:
    GitClient._executable = which('git')



import argparse
import sys

from vcs2l.clients import vcs2l_clients
from vcs2l.commands.command import Command, add_common_arguments
from vcs2l.crawler import find_repositories
from vcs2l.executor import (
    execute_jobs,
    generate_jobs,
    output_repositories,
    output_results,
)
from vcs2l.streams import set_streams


class CustomCommand(Command):
    command = 'custom'
    help = 'Run a custom command'

    def __init__(self, args):
        super(CustomCommand, self).__init__(args)
        self.args = args.args


def get_parser():
    parser = argparse.ArgumentParser(
        description='Run a custom command', prog='vcs custom'
    )
    group = parser.add_argument_group(
        '"custom" command parameters restricting the repositories'
    )
    for client_type in [c.type for c in vcs2l_clients if c.type not in ['tar']]:
        group.add_argument(
            '--' + client_type,
            action='store_true',
            default=False,
            help="Run command on '%s' repositories" % client_type,
        )
    group = parser.add_argument_group('"custom" command parameters')
    group.add_argument(
        '--args',
        required=True,
        nargs='*',
        help='Arbitrary arguments passed '
        'to each vcs invocation. It must be passed after other arguments '
        'since it collects all following options.',
    )
    return parser


def main(args=None, stdout=None, stderr=None):
    set_streams(stdout=stdout, stderr=stderr)

    parser = get_parser()
    add_common_arguments(parser)

    # separate anything followed after --args to not confuse argparse
    if args is None:
        args = sys.argv[1:]
    try:
        index = args.index('--args') + 1
    except ValueError:
        # should generate error due to missing --args
        parser.parse_known_args(args)

    client_args = args[index:]
    args = parser.parse_args(args[0:index])
    args.args = client_args

    # check if any client type is specified
    any_client_type = False
    for client in vcs2l_clients:
        if client.type in args and args.__dict__[client.type]:
            any_client_type = True
            break
    # if no client type is specified enable all client types
    if not any_client_type:
        for client in vcs2l_clients:
            if client.type in args:
                args.__dict__[client.type] = True

    command = CustomCommand(args)

    # filter repositories by specified client types
    clients = find_repositories(command.paths, nested=command.nested)
    clients = [c for c in clients if c.type in args and args.__dict__[c.type]]

    if command.output_repos:
        output_repositories(clients)
    jobs = generate_jobs(clients, command)
    results = execute_jobs(
        jobs, show_progress=True, number_of_workers=args.workers, debug_jobs=args.debug
    )

    output_results(results, hide_empty=args.hide_empty)

    any_error = any(r['returncode'] for r in results)
    return 1 if any_error else 0


def bzr_main(args=None):
    if args is None:
        args = sys.argv[1:]
    return main(['--bzr', '--args'] + args)


def git_main(args=None):
    if args is None:
        args = sys.argv[1:]
    return main(['--git', '--args'] + args)


def hg_main(args=None):
    if args is None:
        args = sys.argv[1:]
    return main(['--hg', '--args'] + args)


def svn_main(args=None):
    if args is None:
        args = sys.argv[1:]
    return main(['--svn', '--args'] + args)


if __name__ == '__main__':
    sys.exit(main())



import argparse
import sys

from vcs2l.commands.command import Command, simple_main
from vcs2l.streams import set_streams


class StatusCommand(Command):
    command = 'status'
    help = 'Show the working tree status'

    def __init__(self, args):
        super(StatusCommand, self).__init__(args)
        self.quiet = args.quiet


def get_parser():
    parser = argparse.ArgumentParser(
        description='Show the working tree status', prog='vcs status'
    )
    group = parser.add_argument_group('"status" command parameters')
    group.add_argument(
        '-q',
        '--quiet',
        action='store_true',
        default=False,
        help="Don't show unversioned items",
    )
    return parser


def main(args=None, stdout=None, stderr=None):
    set_streams(stdout=stdout, stderr=stderr)
    parser = get_parser()
    return simple_main(parser, StatusCommand, args)


if __name__ == '__main__':
    sys.exit(main())



import copy
import os
from shutil import which

from vcs2l.clients.vcs_base import VcsClientBase
from vcs2l.util import rmtree


class BzrClient(VcsClientBase):
    type = 'bzr'
    _executable = None

    @staticmethod
    def is_repository(path):
        return os.path.isdir(os.path.join(path, '.bzr'))

    def __init__(self, path):
        super(BzrClient, self).__init__(path)

    def branch(self, command):
        if command.all:
            return self._not_applicable(
                command, message='at least with the option to list all branches'
            )

        self._check_executable()
        return self._get_parent_branch()

    def custom(self, command):
        self._check_executable()
        cmd = [BzrClient._executable] + command.args
        return self._run_command(cmd)

    def diff(self, _command):
        self._check_executable()
        cmd = [BzrClient._executable, 'diff']
        return self._run_command(cmd)

    def import_(self, command):
        if not command.url:
            return {
                'cmd': '',
                'cwd': self.path,
                'output': "Repository data lacks the 'url' value",
                'returncode': 1,
            }

        self._check_executable()
        if BzrClient.is_repository(self.path):
            # verify that existing repository is the same
            result_parent_branch = self._get_parent_branch()
            if result_parent_branch['returncode']:
                return result_parent_branch
            parent_branch = result_parent_branch['output']
            if parent_branch != command.url:
                if not command.force:
                    return {
                        'cmd': '',
                        'cwd': self.path,
                        'output': 'Path already exists and contains a different '
                        'repository',
                        'returncode': 1,
                    }
                try:
                    rmtree(self.path)
                except OSError:
                    os.remove(self.path)

        not_exist = self._create_path()
        if not_exist:
            return not_exist

        if BzrClient.is_repository(self.path):
            # pull updates for existing repo
            cmd_pull = [BzrClient._executable, 'pull']
            return self._run_command(cmd_pull, retry=command.retry)

        else:
            cmd_branch = [BzrClient._executable, 'branch']
            if command.version:
                cmd_branch += ['-r', command.version]
            cmd_branch += [command.url, '.']
            result_branch = self._run_command(cmd_branch, retry=command.retry)
            if result_branch['returncode']:
                result_branch['output'] = "Could not branch repository '%s': %s" % (
                    command.url,
                    result_branch['output'],
                )
                return result_branch
            return result_branch

    def log(self, command):
        self._check_executable()
        if command.limit_tag or command.limit_untagged:
            tag = None
            if command.limit_tag:
                tag = command.limit_tag
            else:
                # determine nearest tag
                cmd_tag = [BzrClient._executable, 'tags', '--sort=time']
                result_tag = self._run_command(cmd_tag)
                if result_tag['returncode']:
                    return result_tag
                for line in result_tag['output'].splitlines():
                    parts = line.split(' ', 2)
                    if parts[1] != '?':
                        tag = parts[0]
                if not tag:
                    result_tag['output'] = 'Could not determine latest tag'
                    result_tag['returncode'] = 1
                    return result_tag
            # determine revision number of tag
            cmd_tag_rev = [BzrClient._executable, 'revno', '--rev', 'tag:' + tag]
            result_tag_rev = self._run_command(cmd_tag_rev)
            if result_tag_rev['returncode']:
                if command.limit_tag:
                    result_tag_rev['output'] = "Repository lacks the tag '%s'" % tag
                return result_tag_rev
            try:
                tag_rev = int(result_tag_rev['output'])
                tag_next_rev = tag_rev + 1
            except ValueError:
                tag_rev = result_tag_rev['output']
                tag_next_rev = tag_rev
            # determine revision number of HEAD
            cmd_head_rev = [BzrClient._executable, 'revno']
            result_head_rev = self._run_command(cmd_head_rev)
            if result_head_rev['returncode']:
                return result_head_rev
            try:
                head_rev = int(result_head_rev['output'])
            except ValueError:
                head_rev = result_head_rev['output']
            # output log since nearest tag
            cmd_log = [
                BzrClient._executable,
                'log',
                '--rev',
                'revno:%s..' % str(tag_next_rev),
            ]
            if tag_rev == head_rev:
                return {
                    'cmd': ' '.join(cmd_log),
                    'cwd': self.path,
                    'output': '',
                    'returncode': 0,
                }
            if command.limit != 0:
                cmd_log += ['--limit', '%d' % command.limit]
            result_log = self._run_command(cmd_log)
            return result_log
        cmd = [BzrClient._executable, 'log']
        if command.limit != 0:
            cmd += ['--limit', '%d' % command.limit]
        return self._run_command(cmd)

    def pull(self, _command):
        self._check_executable()
        cmd = [BzrClient._executable, 'pull']
        return self._run_command(cmd)

    def push(self, _command):
        self._check_executable()
        cmd = [BzrClient._executable, 'push']
        return self._run_command(cmd)

    def remotes(self, _command):
        self._check_executable()
        return self._get_parent_branch()

    def status(self, _command):
        self._check_executable()
        cmd = [BzrClient._executable, 'status']
        return self._run_command(cmd)

    def _get_parent_branch(self):
        cmd = [BzrClient._executable, 'info']
        # parsing the text output requires enforcing language
        env = copy.copy(os.environ)
        env['LANG'] = 'en_US.UTF-8'
        result = self._run_command(cmd, env)
        if result['returncode']:
            return result
        branch = None
        prefix = '  parent branch: '
        for line in result['output'].splitlines():
            if line.startswith(prefix):
                branch = line[len(prefix) :]
                break
        if not branch:
            result['output'] = ('Could not determine parent branch',)
            result['returncode'] = 1
            return result
        result['output'] = branch
        return result

    def _check_executable(self):
        assert BzrClient._executable is not None, "Could not find 'bzr' executable"


if not BzrClient._executable:
    BzrClient._executable = which('bzr')



import argparse
import sys

from vcs2l.commands.command import Command, simple_main
from vcs2l.streams import set_streams


class BranchCommand(Command):
    command = 'branch'
    help = 'Show the branches'

    def __init__(self, args):
        super(BranchCommand, self).__init__(args)
        self.all = args.all


def get_parser():
    parser = argparse.ArgumentParser(
        description='Show the current branch', prog='vcs branch'
    )
    group = parser.add_argument_group('"branch" command parameters')
    group.add_argument(
        '--all', action='store_true', default=False, help='Show all branches'
    )
    return parser


def main(args=None, stdout=None, stderr=None):
    set_streams(stdout=stdout, stderr=stderr)
    parser = get_parser()
    return simple_main(parser, BranchCommand, args)


if __name__ == '__main__':
    sys.exit(main())



"""Command to delete directories of repositories listed in a YAML file."""

import argparse
import os
import sys
import urllib.request as request
from urllib.error import URLError

from vcs2l.commands.import_ import file_or_url_type, get_repositories
from vcs2l.executor import ansi
from vcs2l.streams import set_streams
from vcs2l.util import rmtree

from .command import Command, existing_dir


class DeleteCommand(Command):
    """Delete directories of repositories listed in a YAML file."""

    command = 'delete'
    help = 'Remove the directories indicated by the list of given repositories.'


def get_parser():
    """CLI parser for the 'delete' command."""
    _cls = DeleteCommand

    parser = argparse.ArgumentParser(
        description=_cls.help, prog='vcs {}'.format(_cls.command)
    )
    group = parser.add_argument_group('Command parameters')
    group.add_argument(
        '--input',
        type=file_or_url_type,
        default='-',
        help='Where to read YAML from',
        metavar='FILE_OR_URL',
    )
    group.add_argument(
        'path',
        nargs='?',
        type=existing_dir,
        default=os.curdir,
        help='Base path to look for repositories',
    )
    group.add_argument(
        '-f',
        '--force',
        action='store_true',
        default=False,
        help='Do the deletion instead of a dry-run',
    )
    return parser


def get_repository_paths(input_source, base_path):
    """Get repository paths from input source."""
    try:
        if isinstance(input_source, request.Request):
            input_source = request.urlopen(input_source)
        repos = get_repositories(input_source)
        return [os.path.join(base_path, rel_path) for rel_path in repos]
    except (RuntimeError, URLError) as e:
        raise RuntimeError(f'Failed to read repositories: {e}') from e


def validate_paths(paths):
    """Validate that paths exist and are directories."""
    valid_paths = []
    missing_paths = []

    for path in paths:
        if os.path.exists(path) and os.path.isdir(path):
            valid_paths.append(path)
        else:
            missing_paths.append(path)

    return valid_paths, missing_paths


def main(args=None, stdout=None, stderr=None):
    """Entry point for the 'delete' command."""

    set_streams(stdout=stdout, stderr=stderr)
    parser = get_parser()
    args = parser.parse_args(args)

    try:
        paths = get_repository_paths(args.input, args.path)
    except RuntimeError as e:
        print(ansi('redf') + f'Error: {e}' + ansi('reset'), file=sys.stderr)
        return 1

    if not paths:
        print(
            ansi('yellowf') + 'No repositories found to delete' + ansi('reset'),
            file=sys.stderr,
        )
        return 0

    # Validate paths existence
    valid_paths, missing_paths = validate_paths(paths)

    if not valid_paths:
        print(
            ansi('redf') + 'No valid directories to delete.' + ansi('reset'),
            file=sys.stderr,
        )
        return 1
    else:
        if missing_paths:
            print(
                ansi('yellowf')
                + 'Warning: The following directories do not exist:'
                + ansi('reset'),
                file=sys.stderr,
            )
            for path in missing_paths:
                print(f'  {path}', file=sys.stderr)

        print(
            ansi('cyanf')
            + 'The following directories will be deleted:'
            + ansi('reset'),
            file=sys.stderr,
        )
        for path in valid_paths:
            print(f'  {path}', file=sys.stderr)

    if not args.force:
        print(
            ansi('yellowf')
            + 'Dry-run mode: No directories were deleted. Use -f/--force to delete them.'
            + ansi('reset'),
            file=sys.stderr,
        )
        return 0

    # Actual deletion
    for path in valid_paths:
        try:
            rmtree(path)
        except OSError as e:
            print(
                ansi('redf') + f'Failed to delete {path}: {e}' + ansi('reset'),
                file=sys.stderr,
            )


if __name__ == '__main__':
    sys.exit(main())



import argparse
import sys

from vcs2l.commands.command import Command, simple_main
from vcs2l.streams import set_streams


class LogCommand(Command):
    command = 'log'
    help = 'Show commit logs'

    def __init__(self, args):
        super(LogCommand, self).__init__(args)
        self.limit = args.limit
        self.limit_tag = args.limit_tag
        self.limit_untagged = args.limit_untagged
        self.merge_only = args.merge_only
        self.verbose = args.verbose


def get_parser():
    parser = argparse.ArgumentParser(description='Show commit logs', prog='vcs log')
    group = parser.add_argument_group('"log" command parameters')
    group.add_argument(
        '-l',
        '--limit',
        metavar='N',
        type=int,
        default=3,
        help='Limit number of logs (0 for unlimited)',
    )
    ex_group = group.add_mutually_exclusive_group()
    ex_group.add_argument(
        '--limit-tag',
        metavar='TAG',
        help='Limit number of log from the head to the specified tag',
    )
    ex_group.add_argument(
        '--limit-untagged',
        action='store_true',
        default=False,
        help='Limit number of log from the head to the last tagged commit',
    )
    group.add_argument(
        '--merge-only',
        action='store_true',
        default=False,
        help='Show only merge commits',
    )
    group.add_argument(
        '--verbose',
        action='store_true',
        default=False,
        help='Show the full commit message',
    )
    return parser


def main(args=None, stdout=None, stderr=None):
    set_streams(stdout=stdout, stderr=stderr)
    parser = get_parser()
    return simple_main(parser, LogCommand, args)


if __name__ == '__main__':
    sys.exit(main())



import os
from shutil import which
from xml.etree.ElementTree import fromstring

from vcs2l.clients.vcs_base import VcsClientBase


class SvnClient(VcsClientBase):
    type = 'svn'
    _executable = None

    @staticmethod
    def is_repository(path):
        return os.path.isdir(os.path.join(path, '.svn'))

    def __init__(self, path):
        super(SvnClient, self).__init__(path)

    def branch(self, command):
        if command.all:
            return self._not_applicable(
                command, message='at least with the option to list all branches'
            )

        self._check_executable()
        cmd_info = [SvnClient._executable, 'info', '--xml']
        result_info = self._run_command(cmd_info)
        if result_info['returncode']:
            result_info['output'] = 'Could not determine url: ' + result_info['output']
            return result_info
        info = result_info['output']

        try:
            root = fromstring(info)
            entry = root.find('entry')
            url = entry.findtext('url')
            repository = entry.find('repository')
            root_url = repository.findtext('root')
        except Exception as e:
            return {
                'cmd': '',
                'cwd': self.path,
                'output': 'Could not determine url from xml: %s' % e,
                'returncode': 1,
            }

        if not url.startswith(root_url):
            return {
                'cmd': '',
                'cwd': self.path,
                'output': "Could not determine url suffix. The root url '%s' is not "
                "a prefix of the url '%s'" % (root_url, url),
                'returncode': 1,
            }

        return {
            'cmd': ' '.join(cmd_info),
            'cwd': self.path,
            'output': url[len(root_url) :],
            'returncode': 0,
        }

    def custom(self, command):
        self._check_executable()
        cmd = [SvnClient._executable] + command.args
        return self._run_command(cmd)

    def diff(self, command):
        self._check_executable()
        cmd = [SvnClient._executable, 'diff']
        if command.context:
            cmd += ['--unified=%d' % command.context]
        return self._run_command(cmd)

    def export(self, command):
        self._check_executable()
        cmd_info = [SvnClient._executable, 'info', '--xml']
        result_info = self._run_command(cmd_info)
        if result_info['returncode']:
            result_info['output'] = 'Could not determine url: ' + result_info['output']
            return result_info
        info = result_info['output']

        try:
            root = fromstring(info)
            entry = root.find('entry')
            url = entry.findtext('url')
            revision = entry.get('revision')
        except Exception as e:
            return {
                'cmd': '',
                'cwd': self.path,
                'output': 'Could not determine url from xml: %s' % e,
                'returncode': 1,
            }

        export_data = {'url': url}
        if command.exact:
            export_data['version'] = revision
        return {
            'cmd': ' '.join(cmd_info),
            'cwd': self.path,
            'output': url,
            'returncode': 0,
            'export_data': export_data,
        }

    def import_(self, command):
        if not command.url:
            return {
                'cmd': '',
                'cwd': self.path,
                'output': "Repository data lacks the 'url' value",
                'returncode': 1,
            }

        not_exist = self._create_path()
        if not_exist:
            return not_exist

        self._check_executable()

        url = command.url
        if command.version:
            url += '@%s' % command.version

        cmd_checkout = [
            SvnClient._executable,
            '--non-interactive',
            'checkout',
            url,
            '.',
        ]
        result_checkout = self._run_command(cmd_checkout, retry=command.retry)
        if result_checkout['returncode']:
            result_checkout['output'] = "Could not checkout repository '%s': %s" % (
                command.url,
                result_checkout['output'],
            )
            return result_checkout

        return {
            'cmd': ' '.join(cmd_checkout),
            'cwd': self.path,
            'output': result_checkout['output'],
            'returncode': 0,
        }

    def log(self, command):
        if command.limit_tag:
            return {
                'cmd': '',
                'cwd': self.path,
                'output': 'SvnClient can not determine log since tag',
                'returncode': NotImplemented,
            }
        if command.limit_untagged:
            return {
                'cmd': '',
                'cwd': self.path,
                'output': 'SvnClient can not determine latest tag',
                'returncode': NotImplemented,
            }
        self._check_executable()
        cmd = [SvnClient._executable, 'log']
        if command.limit != 0:
            cmd += ['--limit', '%d' % command.limit]
        return self._run_command(cmd)

    def pull(self, _command):
        self._check_executable()
        cmd = [SvnClient._executable, '--non-interactive', 'update']
        return self._run_command(cmd)

    def push(self, command):
        self._check_executable()
        return self._not_applicable(command)

    def remotes(self, _command):
        self._check_executable()
        cmd_info = [SvnClient._executable, 'info', '--xml']
        result_info = self._run_command(cmd_info)
        if result_info['returncode']:
            result_info['output'] = 'Could not determine url: ' + result_info['output']
            return result_info
        info = result_info['output']

        try:
            root = fromstring(info)
            entry = root.find('entry')
            url = entry.findtext('url')
        except Exception as e:
            return {
                'cmd': '',
                'cwd': self.path,
                'output': 'Could not determine url from xml: %s' % e,
                'returncode': 1,
            }

        return {
            'cmd': ' '.join(cmd_info),
            'cwd': self.path,
            'output': url,
            'returncode': 0,
        }

    def status(self, command):
        self._check_executable()
        cmd = [SvnClient._executable, 'status']
        if command.quiet:
            cmd += ['--quiet']
        return self._run_command(cmd)

    def validate(self, command):
        if not command.url:
            return {
                'cmd': '',
                'cwd': self.path,
                'output': "Repository data lacks the 'url' value",
                'returncode': 1,
            }

        self._check_executable()

        cmd_info_repo = [SvnClient._executable, 'info', command.url]
        result_info_repo = self._run_command(cmd_info_repo, retry=command.retry)
        if result_info_repo['returncode']:
            result_info_repo['output'] = (
                "Failed to contact remote repository '%s': %s"
                % (command.url, result_info_repo['output'])
            )
            return result_info_repo

        if command.version:
            cmd_info_ver = [
                SvnClient._executable,
                'info',
                command.url + '@' + command.version,
            ]
            result_info_ver = self._run_command(cmd_info_ver, retry=command.retry)

            if result_info_ver['returncode']:
                result_info_ver['output'] = (
                    'Specified version not found on remote repository'
                    + "'%s@%s' : %s"
                    % (command.url, command.version, result_info_ver['output'])
                )
                return result_info_ver

            cmd = result_info_ver['cmd']
            output = "Found svn repository '%s' with revision '%s'" % (
                command.url,
                command.version,
            )
        else:
            cmd = result_info_repo['cmd']
            output = "Found svn repository '%s' with default branch" % command.url

        return {'cmd': cmd, 'cwd': self.path, 'output': output, 'returncode': None}

    def _check_executable(self):
        assert SvnClient._executable is not None, "Could not find 'svn' executable"


if not SvnClient._executable:
    SvnClient._executable = which('svn')



import argparse
import os
import sys
import urllib.request as request
from shutil import which

import yaml

from vcs2l import __version__ as vcs2l_version
from vcs2l.clients import vcs2l_clients
from vcs2l.clients.vcs_base import run_command
from vcs2l.commands.command import Command, add_common_arguments
from vcs2l.executor import ansi, execute_jobs, output_repositories, output_results
from vcs2l.streams import set_streams


class ImportCommand(Command):
    command = 'import'
    help = 'Import the list of repositories'

    def __init__(self, args, url, version=None, recursive=False, shallow=False):
        super(ImportCommand, self).__init__(args)
        self.url = url
        self.version = version
        self.force = args.force
        self.retry = args.retry
        self.skip_existing = args.skip_existing
        self.recursive = recursive
        self.shallow = shallow


def get_parser():
    parser = argparse.ArgumentParser(
        description='Import the list of repositories', prog='vcs import'
    )
    group = parser.add_argument_group('"import" command parameters')
    group.add_argument(
        '--input',
        type=file_or_url_type,
        default='-',
        help='Where to read YAML from',
        metavar='FILE_OR_URL',
    )
    group.add_argument(
        '--force',
        action='store_true',
        default=False,
        help="Delete existing directories if they don't contain the "
        'repository being imported',
    )
    group.add_argument(
        '--shallow',
        action='store_true',
        default=False,
        help='Create a shallow clone without a history',
    )
    group.add_argument(
        '--recursive',
        action='store_true',
        default=False,
        help='Recurse into submodules',
    )
    group.add_argument(
        '--retry',
        type=int,
        metavar='N',
        default=2,
        help='Retry commands requiring network access N times on failure',
    )
    group.add_argument(
        '--skip-existing',
        action='store_true',
        default=False,
        help="Don't overwrite existing directories or change custom checkouts "
        'in repos using the same URL (but fetch repos with same URL)',
    )

    return parser


def file_or_url_type(value):
    if os.path.exists(value) or '://' not in value:
        return argparse.FileType('r')(value)
    # use another user agent to avoid getting a 403 (forbidden) error,
    # since some websites blacklist or block unrecognized user agents
    return request.Request(value, headers={'User-Agent': 'vcs2l/' + vcs2l_version})


def get_repositories(yaml_file):
    try:
        root = yaml.safe_load(yaml_file)
    except yaml.YAMLError as e:
        raise RuntimeError('Input data is not valid yaml format: %s' % e)

    try:
        repositories = root['repositories']
        return get_repos_in_vcs2l_format(repositories)
    except KeyError as e:
        raise RuntimeError('Input data is not valid format: %s' % e)
    except TypeError as e:
        # try rosinstall file format
        try:
            return get_repos_in_rosinstall_format(root)
        except Exception:
            raise RuntimeError('Input data is not valid format: %s' % e)


def get_repos_in_vcs2l_format(repositories):
    repos = {}
    if repositories is None:
        print(
            ansi('yellowf') + 'List of repositories is empty' + ansi('reset'),
            file=sys.stderr,
        )
        return repos
    for path in repositories:
        repo = {}
        attributes = repositories[path]
        if not attributes:
            print(
                ansi('yellowf')
                + "Repository '%s' is empty, skipping" % path
                + ansi('reset'),
                file=sys.stderr,
            )
            continue
        try:
            repo['type'] = attributes['type']
            repo['url'] = attributes['url']
            if 'version' in attributes:
                repo['version'] = attributes['version']
        except KeyError as e:
            print(
                ansi('yellowf')
                + (
                    "Repository '%s' does not provide the necessary "
                    'information: %s' % (path, e)
                )
                + ansi('reset'),
                file=sys.stderr,
            )
            continue
        repos[path] = repo
    return repos


def get_repos_in_rosinstall_format(root):
    repos = {}
    for i, item in enumerate(root):
        if len(item.keys()) != 1:
            raise RuntimeError('Input data is not valid format')
        repo = {'type': list(item.keys())[0]}
        attributes = list(item.values())[0]
        try:
            path = attributes['local-name']
        except KeyError as e:
            print(
                ansi('yellowf')
                + (
                    'Repository #%d does not provide the necessary '
                    'information: %s' % (i, e)
                )
                + ansi('reset'),
                file=sys.stderr,
            )
            continue
        try:
            repo['url'] = attributes['uri']
            if 'version' in attributes:
                repo['version'] = attributes['version']
        except KeyError as e:
            print(
                ansi('yellowf')
                + (
                    "Repository '%s' does not provide the necessary "
                    'information: %s' % (path, e)
                )
                + ansi('reset'),
                file=sys.stderr,
            )
            continue
        repos[path] = repo
    return repos


def generate_jobs(repos, args):
    jobs = []
    for path, repo in repos.items():
        path = os.path.join(args.path, path)
        clients = [c for c in vcs2l_clients if c.type == repo['type']]
        if not clients:
            from vcs2l.clients.none import NoneClient

            job = {
                'client': NoneClient(path),
                'command': None,
                'cwd': path,
                'output': "Repository type '%s' is not supported" % repo['type'],
                'returncode': NotImplemented,
            }
            jobs.append(job)
            continue

        client = clients[0](path)
        command = ImportCommand(
            args,
            repo['url'],
            str(repo['version']) if 'version' in repo else None,
            recursive=args.recursive,
            shallow=args.shallow,
        )
        job = {'client': client, 'command': command}
        jobs.append(job)
    return jobs


def add_dependencies(jobs):
    paths = [job['client'].path for job in jobs]
    for job in jobs:
        job['depends'] = set()
        path = job['client'].path
        while True:
            parent_path = os.path.dirname(path)
            if parent_path == path:
                break
            path = parent_path
            if path in paths:
                job['depends'].add(path)


def main(args=None, stdout=None, stderr=None):
    set_streams(stdout=stdout, stderr=stderr)

    parser = get_parser()
    add_common_arguments(
        parser,
        skip_hide_empty=True,
        skip_nested=True,
        path_nargs='?',
        path_help='Base path to clone repositories to',
    )
    args = parser.parse_args(args)
    try:
        input_ = args.input
        if isinstance(input_, request.Request):
            input_ = request.urlopen(input_)
        repos = get_repositories(input_)
    except (RuntimeError, request.URLError) as e:
        print(ansi('redf') + str(e) + ansi('reset'), file=sys.stderr)
        return 1
    jobs = generate_jobs(repos, args)
    add_dependencies(jobs)

    if args.repos:
        output_repositories([job['client'] for job in jobs])

    workers = args.workers
    # for ssh URLs check if the host is known to prevent ssh asking for
    # confirmation when using more than one worker
    if workers > 1:
        ssh_keygen = None
        checked_hosts = set()
        for job in list(jobs):
            if job['command'] is None:
                continue
            url = job['command'].url
            # only check the host from a ssh URL
            if not url.startswith('git@') or ':' not in url:
                continue
            host = url[4:].split(':', 1)[0]

            # only check each host name once
            if host in checked_hosts:
                continue
            checked_hosts.add(host)

            # get ssh-keygen path once
            if ssh_keygen is None:
                ssh_keygen = which('ssh-keygen') or False
            if not ssh_keygen:
                continue

            result = run_command([ssh_keygen, '-F', host], '')
            if result['returncode']:
                print(
                    'At least one hostname (%s) is unknown, switching to a '
                    'single worker to allow interactively answering the ssh '
                    'question to confirm the fingerprint' % host
                )
                workers = 1
                break

    results = execute_jobs(
        jobs, show_progress=True, number_of_workers=workers, debug_jobs=args.debug
    )
    output_results(results)

    any_error = any(r['returncode'] for r in results)
    return 1 if any_error else 0


if __name__ == '__main__':
    sys.exit(main())



from .branch import BranchCommand
from .custom import CustomCommand
from .delete import DeleteCommand
from .diff import DiffCommand
from .export import ExportCommand
from .import_ import ImportCommand
from .log import LogCommand
from .pull import PullCommand
from .push import PushCommand
from .remotes import RemotesCommand
from .status import StatusCommand
from .validate import ValidateCommand

vcs2l_commands = []
vcs2l_commands.append(BranchCommand)
vcs2l_commands.append(CustomCommand)
vcs2l_commands.append(DeleteCommand)
vcs2l_commands.append(DiffCommand)
vcs2l_commands.append(ExportCommand)
vcs2l_commands.append(ImportCommand)
vcs2l_commands.append(LogCommand)
vcs2l_commands.append(PullCommand)
vcs2l_commands.append(PushCommand)
vcs2l_commands.append(RemotesCommand)
vcs2l_commands.append(StatusCommand)
vcs2l_commands.append(ValidateCommand)

_commands = [c.command for c in vcs2l_commands]
if len(_commands) != len(set(_commands)):
    raise RuntimeError(
        'Multiple commands share the same command name: ' + ', '.join(sorted(_commands))
    )



"""Setup script for the vcs2l package."""

import os

from setuptools import find_packages, setup

from vcs2l import __version__

with open(
    os.path.join(os.path.dirname(__file__), 'README.md'), 'r', encoding='utf-8'
) as f:
    long_description = f.read()

setup(
    name='vcs2l',
    version=__version__,
    requires_python='>=3.6',
    install_requires=[
        "importlib_metadata; python_version<'3.8' and python_version>='3.7'",
        'PyYAML',
        "setuptools; python_version<'3.7'",
    ],
    extras_require={
        'test': [
            'pytest',
        ],
        'docs': [
            'sphinx>=8.2',
            'sphinx-rtd-theme',
        ],
    },
    packages=find_packages(),
    author='Dirk Thomas',
    author_email='web@dirk-thomas.net',
    maintainer='ROS Infrastructure Team',
    project_urls={
        'Source code': 'https://github.com/ros-infrastructure/vcs2l',
        'Issue tracker': 'https://github.com/ros-infrastructure/vcs2l/issues',
    },
    url='https://github.com/ros-infrastructure/vcs2l',
    keywords=['vcs', 'version control', 'git', 'hg', 'svn', 'bzr'],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Programming Language :: Python',
        'Topic :: Software Development :: Version Control',
        'Topic :: Utilities',
    ],
    description='vcs2l provides a command line tool to invoke vcs commands '
    'on multiple repositories.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    license='Apache-2.0',
    data_files=[
        (
            'share/vcs2l-completion',
            [
                'vcs2l-completion/vcs.bash',
                'vcs2l-completion/vcs.tcsh',
                'vcs2l-completion/vcs.zsh',
                'vcs2l-completion/vcs.fish',
            ],
        )
    ],
    entry_points={
        'console_scripts': [
            'vcs = vcs2l.commands.vcs:main',
            'vcs-branch = vcs2l.commands.branch:main',
            'vcs-bzr = vcs2l.commands.custom:bzr_main',
            'vcs-custom = vcs2l.commands.custom:main',
            'vcs-delete = vcs2l.commands.delete:main',
            'vcs-diff = vcs2l.commands.diff:main',
            'vcs-export = vcs2l.commands.export:main',
            'vcs-git = vcs2l.commands.custom:git_main',
            'vcs-help = vcs2l.commands.help:main',
            'vcs-hg = vcs2l.commands.custom:hg_main',
            'vcs-import = vcs2l.commands.import_:main',
            'vcs-log = vcs2l.commands.log:main',
            'vcs-pull = vcs2l.commands.pull:main',
            'vcs-push = vcs2l.commands.push:main',
            'vcs-remotes = vcs2l.commands.remotes:main',
            'vcs-status = vcs2l.commands.status:main',
            'vcs-svn = vcs2l.commands.custom:svn_main',
            'vcs-validate = vcs2l.commands.validate:main',
        ]
    },
)



import os
import subprocess
import sys
import unittest
from shutil import which

sys.path.insert(0, os.path.dirname(os.path.dirname(__file__)))

from vcs2l.clients.git import GitClient  # noqa: E402
from vcs2l.util import rmtree  # noqa: E402

file_uri_scheme = 'file://' if sys.platform != 'win32' else 'file:///'

REPOS_FILE = os.path.join(os.path.dirname(__file__), 'list.repos')
REPOS_FILE_URL = file_uri_scheme + REPOS_FILE
REPOS2_FILE = os.path.join(os.path.dirname(__file__), 'list2.repos')
BAD_REPOS_FILE = os.path.join(os.path.dirname(__file__), 'bad.repos')
TEST_WORKSPACE = os.path.join(
    os.path.dirname(os.path.dirname(__file__)), 'test_workspace'
)

svn = which('svn')
hg = which('hg')
if svn:
    # check if the svn executable is usable (on macOS)
    # and not only exists to state that the program is not installed
    try:
        subprocess.check_call([svn, '--version'])
    except subprocess.CalledProcessError:
        svn = False


class TestCommands(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        assert not os.path.exists(TEST_WORKSPACE)
        os.makedirs(TEST_WORKSPACE)

        try:
            output = run_command('import', ['--input', REPOS_FILE, '.'])
            expected = get_expected_output('import')
            # newer git versions don't append three dots after the commit hash
            assert output == expected or output == expected.replace(b'... ', b' ')
        except Exception:
            cls.tearDownClass()
            raise

    @classmethod
    def tearDownClass(cls):
        rmtree(TEST_WORKSPACE)

    def test_branch(self):
        output = run_command('branch')
        expected = get_expected_output('branch')
        self.assertEqual(output, expected)

    def test_custom(self):
        output = run_command(
            'custom',
            args=['--git', '--args', 'describe', '--abbrev=0', '--tags'],
            subfolder='immutable',
        )
        expected = get_expected_output('custom_describe')
        self.assertEqual(output, expected)

    def test_diff(self):
        license_path = os.path.join(TEST_WORKSPACE, 'immutable', 'hash', 'LICENSE')
        file_length = None
        try:
            with open(license_path, 'ab') as h:
                file_length = h.tell()
                h.write(b'testing')

            output = run_command('diff', args=['--hide'])
            expected = get_expected_output('diff_hide')
        finally:
            if file_length is not None:
                with open(license_path, 'ab') as h:
                    h.truncate(file_length)

        self.assertEqual(output, expected)

    def test_export_exact_with_tags(self):
        output = run_command(
            'export', args=['--exact-with-tags'], subfolder='immutable'
        )
        expected = get_expected_output('export_exact_with_tags')
        self.assertEqual(output, expected)

    def test_export_exact(self):
        output = run_command('export', args=['--exact'], subfolder='immutable')
        expected = get_expected_output('export_exact')
        self.assertEqual(output, expected)

    def test_log(self):
        output = run_command('log', args=['--limit', '2'], subfolder='immutable')
        expected = get_expected_output('log_limit')
        self.assertEqual(output, expected)

    def test_log_merge_only(self):
        output = run_command('log', args=['--merge-only'], subfolder='immutable/tag')
        expected = get_expected_output('log_merges_only')
        self.assertEqual(output, expected)

    def test_pull(self):
        output = run_command('pull', args=['--workers', '1'])
        expected = get_expected_output('pull')
        # replace message from older git versions
        output = output.replace(
            b'anch. Please specify which\nbranch you want to merge with. See',
            b'anch.\nPlease specify which branch you want to merge with.\nSee',
        )
        self.assertEqual(output, expected)

    def test_pull_api(self):
        from io import StringIO

        from vcs2l.commands.pull import main

        stdout_stderr = StringIO()

        # change and restore cwd
        cwd_bck = os.getcwd()
        os.chdir(TEST_WORKSPACE)
        try:
            # change and restore USE_COLOR flag
            from vcs2l import executor

            use_color_bck = executor.USE_COLOR
            executor.USE_COLOR = False
            try:
                # change and restore os.environ
                env_bck = os.environ
                os.environ = dict(os.environ)
                os.environ.update(
                    LANG='en_US.UTF-8',
                    PYTHONPATH=(
                        os.path.dirname(os.path.dirname(__file__))
                        + os.pathsep
                        + os.environ.get('PYTHONPATH', '')
                    ),
                )
                try:
                    rc = main(
                        args=['--workers', '1'],
                        stdout=stdout_stderr,
                        stderr=stdout_stderr,
                    )
                finally:
                    os.environ = env_bck
            finally:
                executor.USE_COLOR = use_color_bck
        finally:
            os.chdir(cwd_bck)

        assert rc == 0
        # replace message from older git versions
        output = stdout_stderr.getvalue().replace(
            'anch. Please specify which\nbranch you want to merge with. See',
            'anch.\nPlease specify which branch you want to merge with.\nSee',
        )
        # the output was retrieved through a different way here
        output = adapt_command_output(output.encode()).decode()
        if sys.platform == 'win32':
            # it does not include carriage return characters on Windows
            output = output.replace('\n', '\r\n')
        expected = get_expected_output('pull').decode()
        assert output == expected

    def test_reimport(self):
        cwd_vcs2l = os.path.join(TEST_WORKSPACE, 'vcs2l')
        subprocess.check_output(
            ['git', 'remote', 'add', 'foo', 'http://foo.com/bar.git'],
            stderr=subprocess.STDOUT,
            cwd=cwd_vcs2l,
        )
        cwd_without_version = os.path.join(TEST_WORKSPACE, 'without_version')
        subprocess.check_output(
            ['git', 'checkout', '-b', 'foo'],
            stderr=subprocess.STDOUT,
            cwd=cwd_without_version,
        )
        output = run_command('import', ['--skip-existing', '--input', REPOS_FILE, '.'])
        expected = get_expected_output('reimport_skip')
        # newer git versions don't append three dots after the commit hash
        assert output == expected or output == expected.replace(b'... ', b' ')

        subprocess.check_output(
            ['git', 'remote', 'set-url', 'origin', 'http://foo.com/bar.git'],
            stderr=subprocess.STDOUT,
            cwd=cwd_without_version,
        )
        run_command('import', ['--skip-existing', '--input', REPOS_FILE, '.'])

        output = run_command('import', ['--force', '--input', REPOS_FILE, '.'])
        expected = get_expected_output('reimport_force')
        # on Windows, the "Already on 'main'" message is after the
        # "Your branch is up to date with ..." message, so remove it
        # from both output and expected strings
        if sys.platform == 'win32':
            output = output.replace(b"Already on 'main'\r\n", b'')
            expected = expected.replace(b"Already on 'main'\r\n", b'')
        # newer git versions don't append three dots after the commit hash
        assert output == expected or output == expected.replace(b'... ', b' ')

        subprocess.check_output(
            ['git', 'remote', 'remove', 'foo'], stderr=subprocess.STDOUT, cwd=cwd_vcs2l
        )

    def test_reimport_failed(self):
        cwd_tag = os.path.join(TEST_WORKSPACE, 'immutable', 'tag')
        subprocess.check_output(
            ['git', 'remote', 'add', 'foo', 'http://foo.com/bar.git'],
            stderr=subprocess.STDOUT,
            cwd=cwd_tag,
        )
        subprocess.check_output(
            ['git', 'remote', 'rm', 'origin'], stderr=subprocess.STDOUT, cwd=cwd_tag
        )
        try:
            run_command('import', ['--skip-existing', '--input', REPOS_FILE, '.'])
        finally:
            subprocess.check_output(
                ['git', 'remote', 'rm', 'foo'], stderr=subprocess.STDOUT, cwd=cwd_tag
            )
            subprocess.check_output(
                [
                    'git',
                    'remote',
                    'add',
                    'origin',
                    'https://github.com/ros-infrastructure/vcs2l.git',
                ],
                stderr=subprocess.STDOUT,
                cwd=cwd_tag,
            )

    def test_import_force_non_empty(self):
        workdir = os.path.join(TEST_WORKSPACE, 'force-non-empty')
        os.makedirs(os.path.join(workdir, 'vcs2l', 'not-a-git-repo'))
        try:
            output = run_command(
                'import',
                ['--force', '--input', REPOS_FILE, '.'],
                subfolder='force-non-empty',
            )
            expected = get_expected_output('import')
            # newer git versions don't append ... after the commit hash
            assert output == expected or output == expected.replace(b'... ', b' ')
        finally:
            rmtree(workdir)

    def test_import_shallow(self):
        workdir = os.path.join(TEST_WORKSPACE, 'import-shallow')
        os.makedirs(workdir)
        try:
            output = run_command(
                'import',
                ['--shallow', '--input', REPOS_FILE, '.'],
                subfolder='import-shallow',
            )
            # the actual output contains absolute paths
            output = output.replace(
                b'repository in ' + workdir.encode() + b'/', b'repository in ./'
            )
            expected = get_expected_output('import_shallow')
            # newer git versions don't append ... after the commit hash
            assert output == expected or output == expected.replace(b'... ', b' ')

            # check that repository history has only one commit
            output = subprocess.check_output(
                ['git', 'log', '--format=oneline'],
                stderr=subprocess.STDOUT,
                cwd=os.path.join(workdir, 'vcs2l'),
            )
            assert len(output.splitlines()) == 1
        finally:
            rmtree(workdir)

    def test_import_url(self):
        workdir = os.path.join(TEST_WORKSPACE, 'import-url')
        os.makedirs(workdir)
        try:
            output = run_command(
                'import', ['--input', REPOS_FILE_URL, '.'], subfolder='import-url'
            )
            # the actual output contains absolute paths
            output = output.replace(
                b'repository in ' + workdir.encode() + b'/', b'repository in ./'
            )
            expected = get_expected_output('import')
            # newer git versions don't append ... after the commit hash
            assert output == expected or output == expected.replace(b'... ', b' ')
        finally:
            rmtree(workdir)

    def test_deletion(self):
        """Test the delete command."""
        workdir = os.path.join(TEST_WORKSPACE, 'deletion')
        os.makedirs(workdir)
        try:
            run_command(
                'import', ['--input', REPOS_FILE_URL, '.'], subfolder='deletion'
            )
            output = run_command(
                'delete',
                ['--force', '--input', REPOS_FILE_URL, '.'],
                subfolder='deletion',
            )
            expected = get_expected_output('delete')
            # we don't care what order these messages appear in
            output = b'\n'.join(sorted(output.split(b'\n')))
            expected = b'\n'.join(sorted(output.split(b'\n')))
            self.assertEqual(output, expected)

            # check that repositories were actually deleted
            self.assertFalse(os.path.exists(os.path.join(workdir, 'immutable/hash')))
            self.assertFalse(
                os.path.exists(os.path.join(workdir, 'immutable/hash_tar'))
            )
            self.assertFalse(
                os.path.exists(os.path.join(workdir, 'immutable/hash_zip'))
            )
            self.assertFalse(os.path.exists(os.path.join(workdir, 'immutable/tag')))
            self.assertFalse(os.path.exists(os.path.join(workdir, 'vcs2l')))
            self.assertFalse(os.path.exists(os.path.join(workdir, 'without_version')))
        finally:
            rmtree(workdir)

    def test_validate(self):
        output = run_command('validate', ['--input', REPOS_FILE])
        expected = get_expected_output('validate')
        self.assertEqual(output, expected)

        output = run_command('validate', ['--hide-empty', '--input', REPOS_FILE])
        expected = get_expected_output('validate_hide')
        # we don't care what order these messages appear in
        output = b'\n'.join(sorted(output.split(b'\n')))
        expected = b'\n'.join(sorted(output.split(b'\n')))
        self.assertEqual(output, expected)

        output = run_command('validate', ['--input', BAD_REPOS_FILE])
        expected = get_expected_output('validate_bad')
        self.assertEqual(output, expected)

    @unittest.skipIf(not svn, '`svn` was not found')
    @unittest.skipIf(not hg, '`hg` was not found')
    def test_validate_svn_and_hg(self):
        output = run_command('validate', ['--input', REPOS2_FILE])
        expected = get_expected_output('validate2')
        self.assertEqual(output, expected)

    def test_remote(self):
        output = run_command('remotes', args=['--repos'])
        expected = get_expected_output('remotes_repos')
        self.assertEqual(output, expected)

    def test_status(self):
        output = run_command('status')
        # replace message from older git versions
        # https://github.com/git/git/blob/3ec7d702a89c647ddf42a59bc3539361367de9d5/Documentation/RelNotes/2.10.0.txt#L373-L374
        output = output.replace(b'working directory clean', b'working tree clean')
        # the following seems to have changed between git 2.10.0 and 2.14.1
        output = output.replace(b'.\nnothing to commit', b'.\n\nnothing to commit')
        expected = get_expected_output('status')
        self.assertEqual(output, expected)


def run_command(command, args=None, subfolder=None):
    repo_root = os.path.dirname(os.path.dirname(__file__))
    script = os.path.join(repo_root, 'scripts', 'vcs-' + command)
    env = dict(os.environ)
    env.update(
        GIT_CONFIG_GLOBAL=os.path.join(repo_root, 'test', '.gitconfig'),
        LANG='en_US.UTF-8',
        PYTHONPATH=repo_root + os.pathsep + env.get('PYTHONPATH', ''),
        PYTHONWARNINGS='ignore',
    )
    cwd = TEST_WORKSPACE
    if subfolder:
        cwd = os.path.join(cwd, subfolder)
    output = subprocess.check_output(
        [sys.executable, script] + (args or []),
        stderr=subprocess.STDOUT,
        cwd=cwd,
        env=env,
    )
    return adapt_command_output(output, cwd)


def adapt_command_output(output, cwd=None):
    assert isinstance(output, bytes)
    # replace message from older git versions
    output = output.replace(
        b'git checkout -b new_branch_name', b'git checkout -b <new-branch-name>'
    )
    output = output.replace(b'(detached from ', b'(HEAD detached at ')
    output = output.replace(
        b"ady on 'main'\n=",
        b"ady on 'main'\nYour branch is up-to-date with 'origin/main'.\n=",
    )
    output = output.replace(b'# HEAD detached at ', b'HEAD detached at ')
    output = output.replace(
        b'# On branch main',
        b"On branch main\nYour branch is up-to-date with 'origin/main'.\n",
    )
    # the following seems to have changed between git 2.17.1 and 2.25.1
    output = output.replace(b"Note: checking out '", b"Note: switching to '")
    output = output.replace(
        b'by performing another checkout.', b'by switching back to a branch.'
    )
    output = output.replace(
        b'using -b with the checkout command again.',
        b'using -c with the switch command.',
    )
    output = output.replace(
        b'git checkout -b <new-branch-name>',
        b'git switch -c <new-branch-name>\n\n'
        b'Or undo this operation with:\n\n'
        b'  git switch -\n\n'
        b'Turn off this advice by setting config variable '
        b'advice.detachedHead to false',
    )
    if sys.platform == 'win32':
        if cwd:
            # on Windows, git prints full path to repos
            # in some messages, so make it relative
            cwd_abs = os.path.abspath(cwd).replace('\\', '/')
            output = output.replace(cwd_abs.encode(), b'.')
        # replace path separators in specific paths;
        # this is less likely to cause wrong test results
        paths_to_replace = [
            (b'.\\immutable', b'./immutable'),
            (b'.\\vcs2l', b'./vcs2l'),
            (b'.\\without_version', b'./without_version'),
            (b'\\hash', b'/hash'),
            (b'\\tag', b'/tag'),
        ]
        for before, after in paths_to_replace:
            output = output.replace(before, after)
    return output


def get_expected_output(name):
    path = os.path.join(os.path.dirname(__file__), name + '.txt')
    with open(path, 'rb') as h:
        content = h.read()
    # change in git version 2.15.0
    # https://github.com/git/git/commit/7560f547e6
    if GitClient.get_git_version() < [2, 15, 0]:
        # use hyphenation for older git versions
        content = content.replace(b'up to date', b'up-to-date')
    return content


if __name__ == '__main__':
    unittest.main()



import argparse
import os
import sys

from vcs2l.commands.command import Command, add_common_arguments
from vcs2l.crawler import find_repositories
from vcs2l.executor import (
    ansi,
    execute_jobs,
    generate_jobs,
    output_repositories,
    output_results,
)
from vcs2l.streams import set_streams


class ExportCommand(Command):
    command = 'export'
    help = 'Export the list of repositories'

    def __init__(self, args):
        super(ExportCommand, self).__init__(args)
        self.exact = args.exact or args.exact_with_tags
        self.with_tags = args.exact_with_tags


def get_parser():
    parser = argparse.ArgumentParser(
        description='Export the list of repositories', prog='vcs export'
    )
    group = parser.add_argument_group('"export" command parameters')
    group_exact = group.add_mutually_exclusive_group()
    group_exact.add_argument(
        '--exact',
        action='store_true',
        default=False,
        help='Export commit hashes instead of branch names',
    )
    group_exact.add_argument(
        '--exact-with-tags',
        action='store_true',
        default=False,
        help='Export unique tag names or commit hashes instead of branch names',
    )
    group.add_argument(
        '--lint',
        action='store_true',
        default=False,
        help='Format output for compatibility with yamllint',
    )

    return parser


def output_export_data(result, hide_empty=False):
    # errors are handled by a separate function
    if result['returncode']:
        return

    try:
        lines = []
        lines.append('  %s:' % result['path'])
        lines.append('    type: ' + result['client'].__class__.type)
        export_data = result['export_data']
        lines.append('    url: ' + export_data['url'])
        if 'version' in export_data and export_data['version']:
            lines.append('    version: ' + export_data['version'])
        print('\n'.join(lines))
    except KeyError as e:
        print(
            ansi('redf')
            + (
                "Command '%s' failed for path '%s': %s: %s"
                % (
                    result['command'].__class__.command,
                    result['client'].path,
                    e.__class__.__name__,
                    e,
                )
            )
            + ansi('reset'),
            file=sys.stderr,
        )


def output_error_information(result, hide_empty=False):
    # successful results are handled by a separate function
    if not result['returncode']:
        return

    if result['returncode'] == NotImplemented:
        color = 'yellow'
    else:
        color = 'red'

    line = '%s: %s' % (result['path'], result['output'])
    print(ansi('%sf' % color) + line + ansi('reset'), file=sys.stderr)


def get_relative_path_of_result(result):
    client = result['client']
    return os.path.relpath(client.path, result['command'].paths[0])


def main(args=None, stdout=None, stderr=None):
    set_streams(stdout=stdout, stderr=stderr)

    parser = get_parser()
    add_common_arguments(parser, skip_hide_empty=True, path_nargs='?')
    args = parser.parse_args(args)

    command = ExportCommand(args)
    clients = find_repositories(command.paths, nested=command.nested)
    if command.output_repos:
        output_repositories(clients)
    jobs = generate_jobs(clients, command)
    results = execute_jobs(jobs, number_of_workers=args.workers)

    # check if at least one repo was found in the client directory
    basename = None
    for result in results:
        result['path'] = get_relative_path_of_result(result)
        if result['path'] == '.':
            basename = os.path.basename(os.path.abspath(result['client'].path))
    # in that case prefix all relative paths with the client directory basename
    if basename is not None:
        for result in results:
            if result['path'] == '.':
                result['path'] = basename
            else:
                result['path'] = os.path.join(basename, result['path'])

    if args.lint:
        print('---')
    print('repositories:')
    output_results(results, output_handler=output_export_data)
    output_results(results, output_handler=output_error_information)
    if args.lint:
        print('...')

    any_error = any(r['returncode'] for r in results)
    return 1 if any_error else 0


if __name__ == '__main__':
    sys.exit(main())



import os
import socket
import subprocess
import sys
import time
from urllib.error import HTTPError, URLError
from urllib.request import Request, urlopen

from vcs2l.executor import ansi


class VcsClientBase(object):
    type = None

    def __init__(self, path):
        self.path = path

    def __getattribute__(self, name):
        if name == 'import':
            try:
                return self.import_
            except AttributeError:
                pass
        return super(VcsClientBase, self).__getattribute__(name)

    def _not_applicable(self, command, message=None):
        return {
            'cmd': '%s.%s(%s)'
            % (self.__class__.type, 'push', command.__class__.command),
            'output': "Command '%s' not applicable for client '%s'%s"
            % (
                command.__class__.command,
                self.__class__.type,
                ': ' + message if message else '',
            ),
            'returncode': NotImplemented,
        }

    def _run_command(self, cmd, env=None, retry=0):
        for i in range(retry + 1):
            if i > 0:
                print(
                    ansi('yellowf')
                    + 'Retrying command (%d/%d): %s' % (i, retry, ' '.join(cmd))
                    + ansi('reset'),
                    file=sys.stderr,
                )
            result = run_command(cmd, os.path.abspath(self.path), env=env)
            if not result['returncode']:
                # return successful result
                break
            if i >= retry:
                # return the failure after retries
                break
            # increasing sleep before each retry
            time.sleep(i + 1)
        return result

    def _create_path(self):
        if not os.path.exists(self.path):
            try:
                os.makedirs(self.path)
            except os.error as e:
                return {
                    'cmd': 'os.makedirs(%s)' % self.path,
                    'cwd': self.path,
                    'output': "Could not create directory '%s': %s" % (self.path, e),
                    'returncode': 1,
                }
        return None


def run_command(cmd, cwd, env=None):
    if not os.path.exists(cwd):
        cwd = None
    result = {'cmd': ' '.join(cmd), 'cwd': cwd}
    try:
        proc = subprocess.Popen(
            cmd, cwd=cwd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, env=env
        )
        output, _ = proc.communicate()
        result['output'] = output.rstrip().decode('utf8')
        result['returncode'] = proc.returncode
    except subprocess.CalledProcessError as e:
        result['output'] = e.output.decode('utf8')
        result['returncode'] = e.returncode
    return result


def load_url(url, retry=2, retry_period=1, timeout=10):
    try:
        fh = urlopen(url, timeout=timeout)
    except HTTPError as e:
        if e.code == 503 and retry:
            time.sleep(retry_period)
            return load_url(
                url, retry=retry - 1, retry_period=retry_period, timeout=timeout
            )
        e.msg += ' (%s)' % url
        raise
    except URLError as e:
        if isinstance(e.reason, socket.timeout) and retry:
            time.sleep(retry_period)
            return load_url(
                url, retry=retry - 1, retry_period=retry_period, timeout=timeout
            )
        raise URLError(str(e) + ' (%s)' % url)
    return fh.read()


def test_url(url, retry=2, retry_period=1, timeout=10):
    request = Request(url)
    request.get_method = lambda: 'HEAD'

    try:
        response = urlopen(request)
    except HTTPError as e:
        if e.code == 503 and retry:
            time.sleep(retry_period)
            return test_url(
                url, retry=retry - 1, retry_period=retry_period, timeout=timeout
            )
        e.msg += ' (%s)' % url
        raise
    except URLError as e:
        if isinstance(e.reason, socket.timeout) and retry:
            time.sleep(retry_period)
            return test_url(
                url, retry=retry - 1, retry_period=retry_period, timeout=timeout
            )
        raise URLError(str(e) + ' (%s)' % url)
    return response



import os
import tarfile
from io import BytesIO
from urllib.error import URLError

from vcs2l.clients.vcs_base import VcsClientBase, load_url, test_url
from vcs2l.util import rmtree


class TarClient(VcsClientBase):
    type = 'tar'

    @staticmethod
    def is_repository(path):
        return False

    def __init__(self, path):
        super(TarClient, self).__init__(path)

    def import_(self, command):
        if not command.url:
            return {
                'cmd': '',
                'cwd': self.path,
                'output': "Repository data lacks the 'url' value",
                'returncode': 1,
            }

        # clear destination
        if os.path.exists(self.path):
            for filename in os.listdir(self.path):
                path = os.path.join(self.path, filename)
                try:
                    rmtree(path)
                except OSError:
                    os.remove(path)
        else:
            not_exist = self._create_path()
            if not_exist:
                return not_exist

        # download tarball
        try:
            data = load_url(command.url, retry=command.retry)
        except URLError as e:
            return {
                'cmd': '',
                'cwd': self.path,
                'output': "Could not fetch tarball from '%s': %s" % (command.url, e),
                'returncode': 1,
            }

        # unpack tarball into destination
        try:
            # raise all fatal errors
            tar = tarfile.open(mode='r', fileobj=BytesIO(data), errorlevel=1)
        except (tarfile.ReadError, IOError, OSError) as e:
            return {
                'cmd': '',
                'cwd': self.path,
                'output': "Failed to read tarball fetched from '%s': %s"
                % (command.url, e),
                'returncode': 1,
            }

        if not command.version:
            members = None
        else:
            # remap all members from version subfolder into destination
            def get_members(tar, prefix):
                for tar_info in tar.getmembers():
                    if tar_info.name.startswith(prefix):
                        tar_info.name = tar_info.name[len(prefix) :]
                        yield tar_info

            prefix = str(command.version) + '/'
            members = get_members(tar, prefix)

        if hasattr(tarfile, 'data_filter'):
            tar.extractall(self.path, members, filter='data')
        else:
            tar.extractall(self.path, members)

        return {
            'cmd': '',
            'cwd': self.path,
            'output': "Downloaded tarball from '%s' and unpacked it" % command.url,
            'returncode': 0,
        }

    def validate(self, command):
        if not command.url:
            return {
                'cmd': '',
                'cwd': self.path,
                'output': "Repository data lacks the 'url' value",
                'returncode': 1,
            }

        # test url
        try:
            test_url(command.url, retry=command.retry)
        except URLError as e:
            return {
                'cmd': '',
                'cwd': self.path,
                'output': "Failed to contact tarball url '%s': %s" % (command.url, e),
                'returncode': 1,
            }
        return {
            'cmd': 'http HEAD url',
            'cwd': self.path,
            'output': "Tarball url '%s' exists" % command.url,
            'returncode': None,
        }



import os

from . import vcs2l_clients


def find_repositories(paths, nested=False):
    repos = []
    visited = []
    for path in paths:
        _find_repositories(path, repos, visited, nested=nested)
    return repos


def _find_repositories(path, repos, visited, nested=False):
    abs_path = os.path.abspath(path)
    if abs_path in visited:
        return
    visited.append(abs_path)

    client = get_vcs_client(path)
    if client:
        repos.append(client)
        if not nested:
            return

    try:
        listdir = os.listdir(path)
    except OSError:
        listdir = []
    for name in sorted(listdir):
        subpath = os.path.join(path, name)
        if not os.path.isdir(subpath):
            continue
        _find_repositories(subpath, repos, visited, nested=nested)


def get_vcs_client(path):
    for client_class in vcs2l_clients:
        if client_class.is_repository(path):
            return client_class(path)
    return None



import sys

from vcs2l.commands.help import get_entrypoint, get_parser
from vcs2l.commands.help import main as help_main
from vcs2l.streams import set_streams


def main(args=None, stdout=None, stderr=None):
    set_streams(stdout=stdout, stderr=stderr)

    # no help to extract command first (which might be followed by --help)
    parser = get_parser(add_help=False)
    ns, _ = parser.parse_known_args(args)
    args = args if args is not None else sys.argv[1:]

    # relay to specific command
    if ns.command and ns.command != 'help':
        entrypoint = get_entrypoint(ns.command)
        if not entrypoint:
            return 1

        args.remove(ns.command)
        return entrypoint(args)

    # remove help command if specified
    if ns.command:
        args.remove(ns.command)

    return help_main(args)


if __name__ == '__main__':
    sys.exit(main())
